#include "config.h"

#ifdef USE_MPI
#include <mpi.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include "CWave.h"
#include "display.h"
#include "memory.h"
#include "mesh.h"

/*********************************************************************************************
 * TVD limiter added by Bob Robey 2010
 *
 * WAVE -- 2D Shallow Water Equation Model
 *                            Bob Robey, Los Alamos National Laboratory
 *                      Written for Supercomputing Challenge Kickoff 2007
 *                                   Copyright, 2007
 *                      Permission Granted to Supercomputing Challenge
 *                           Participants to use in their projects
 
 * This work is based on the following 2006-2007 Supercomputing
 * Challenge Projects. Their work is used by permission.
 *  
 * Numerical method based on the Shallow Water Model by
 * 
 *             Joseph Koby, Sarah Armstrong, Juan-Antonio Vigil, Vanessa Trujillo,
 *                                  McCurdy High School
 *                                    Copyright 2007
 *                        Written for Supercomputing Challenge 2006-2007
 *                                ALL RIGHTS RESERVED
 * 
 * C/MPI coding based on the SAPIENT Compressible Fluid Model
 * 
 *                                      SAPIENT
 *                  Copyright 2007 Jonathan Robey, Dov Shlachter
 *                               Los Alamos High School
 *                     Written for Supercomputing Challenge 2006-2007
 *                                  ALL RIGHTS RESERVED
 * 
 * ******************************************************************************************/

//define macro for squaring a number
#define SQ(x) ((x)*(x))
//define macro to find maximum of 3 values
#define MAX3(a,b,c) (max(max((a),(b)),(c)))
//define macro to find maximum of 2 values
#define max(f,g) ((f)>(g))?(f):(g)
//define macro to find minimum of 2 values
#define min(f,g) ((f)<(g))?(f):(g)
//define macro to find minimum of 3 values
#define MIN3(a,b,c) (min(min((a),(b)),(c)))
//define macro for swapping pointers
#define SWAP(xnew,xold,xtmp) (xtmp=xnew, xnew=xold, xold=xtmp)

#define HXFLUX(j,i)  ( U[j+2][i+2] )
#define UXFLUX(j,i)  ( SQ(U[j+2][i+2])/H[j+2][i+2] + ghalf*SQ(H[j+2][i+2]) )
#define UVFLUX(j,i)  ( U[j+2][i+2]*V[j+2][i+2]/H[j+2][i+2] )

#define HYFLUX(j,i)  ( V[j+2][i+2] )
#define VUFLUX(j,i)  ( V[j+2][i+2]*U[j+2][i+2]/H[j+2][i+2] )
#define VYFLUX(j,i)  ( SQ(V[j+2][i+2])/H[j+2][i+2] + ghalf*SQ(H[j+2][i+2]) )

#define HNEWXFLUX(j,i)  ( Ux[j-2][i-2] )
#define UNEWXFLUX(j,i)  ( SQ(Ux[j-2][i-2])/Hx[j-2][i-2] + ghalf*SQ(Hx[j-2][i-2]) )
#define UVNEWFLUX(j,i)  ( Ux[j-2][i-2]*Vx[j-2][i-2]/Hx[j-2][i-2] )

#define HNEWYFLUX(j,i)  ( Vy[j-2][i-2] )
#define VNEWYFLUX(j,i)  ( SQ(Vy[j-2][i-2])/Hy[j-2][i-2] + ghalf*SQ(Hy[j-2][i-2]) )
#define VUNEWFLUX(j,i)  ( Vy[j-2][i-2]*Uy[j-2][i-2]/Hy[j-2][i-2] )

double   **H, **U, **V;                 //state variables
double   **Hnew, **Unew, **Vnew;        //
double   **Hx, **Ux, **Vx;  //half-step arrays
double   **Hy, **Uy, **Vy;  //half-step arrays

void mesh_allocation(unsigned int imax, unsigned int mysize);

int main(int argc, char *argv[])
{
  int rank, numpes;
  unsigned int i, j;
  unsigned int imax, jmax;
  unsigned int ntimes;
  int ndisplay = 100;
  unsigned int n;
  unsigned int mysize, myoffset;
  unsigned int meshRow=2;
  int ibutton = 0;
  int special_event;
  int icell, jcell;
  char key='\0';
  int xloc, yloc;
  int runflag = 1;
  double xcor, ycor;
  double   deltaX, deltaY;         //size of cell
  double   g = 9.80;               // gravitational constant
  double   ghalf = 0.5*g;
  double   deltaT = 0.025;         //default timestep
  double   sigma = 0.9;
  double   colorScale=10.0;
  double   graphScale=11.0;
  double   time=0.0;               //computer simulation time
  double   localtime, totaltime, elapsed_time;   //variables to calculate time taken for the program to run
  double   localtime_reduction=0.0, elapsed_time_reduction;
  struct timeval  tstart, tstop, tresult;
  struct timeval  tstart_sum, tstop_sum;
  double   TotalMass, origTM;            //variables for checking conservation of mass
  char     string[80];
  double duminus1, duminus2;
  double duplus1,  duplus2;
  double duhalf1,  duhalf2;
  double rdenom, rnumplus, rnumminus, rminus, rplus;
  double q, nu, cv, wminusx, wplusx, wminusy, wplusy;
  double **tmp;
  double **display_var;
  double Cx, Cy, Cxhalf, Cyhalf;
    
#ifdef USE_MPI
  MPI_Init (&argc, &argv);
  sum_init();
  
  /* Determine size and my rank in MPI_COMM_WORLD communicator */
  
  MPI_Comm_size(MPI_COMM_WORLD, &numpes) ;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank) ;    
#endif

  if(DEBUG>=1) printf("My rank is %d and mysize is %d\n", rank, mysize);

  if(rank == 0) printf("Copyright 2007\n");
 
  /* If I'm process 0, determine the matrix size and # of iterations*/
  if (rank == 0) {
      /*
      printf("Matrix Size X : ");
      scanf("%d",&imax);
      printf("Matrix Size Y : ") ;
      scanf("%d",&jmax) ;
      printf("Iterations : ") ;
      scanf("%d",&ntimes) ;//*/
      imax = 1280; jmax = 1280; ntimes = 10000;
      //imax = 2048; jmax = 2048; ntimes = 1000;
  }
  
#ifdef USE_MPI
   /* Broadcast the size and # of iterations to all processes */
   MPI_Bcast(&imax, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
   MPI_Bcast(&jmax, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
   MPI_Bcast(&ntimes, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif

   /* Set neighbors and determine my part of the matrix */
   mesh_set_size(imax, jmax);
   mysize = mesh_get_mysize();
   myoffset = mesh_get_myoffset();

   /* allocate the memory dynamically for the matrix */
   mesh_allocation(imax, mysize);

   if(rank==0&&DEBUG>=1) printf("Memory allocated\n");

   /*initialize matrix*/
   mesh_initialize(PROBLEM, H, U, V);
   deltaX = mesh_get_deltaX();
   deltaY = mesh_get_deltaY();
   double xmin = mesh_get_xmin();
   double xmax = mesh_get_xmax();
   double ymin = mesh_get_ymin();
   double ymax = mesh_get_ymax();
   double *x = mesh_get_x();
   double *y = mesh_get_y();
   double *dx = mesh_get_dx();
   double *dy = mesh_get_dy();

  gettimeofday(&tstart_sum, NULL);
  origTM = sum_var(H);
  gettimeofday(&tstop_sum, NULL);
  timersub(&tstop_sum,&tstart_sum,&tresult);
  localtime_reduction += tresult.tv_sec + tresult.tv_usec/1000000.0;

  // We calculate a timestep based on SQ(c)=g*h and not let a wave
  // cross a cell.
  gettimeofday(&tstart_sum, NULL);
  deltaT = mesh_set_timestep(g, sigma, H, U, V);
  gettimeofday(&tstop_sum, NULL);
  timersub(&tstop_sum,&tstart_sum,&tresult);
  localtime_reduction += tresult.tv_sec + tresult.tv_usec/1000000.0;

  if(rank==0&&DEBUG>=1) printf("initial values set\n");
  //display initial values

  if (DEBUG >=2){
    mesh_output("Init", 0, imax+4, 0, mysize+4, H, U, V);
  }

  if (DISPLAY_ON) {
     display_init(&argc, &argv, 800, 800, 600, xmin, xmax, ymin, ymax);
     display_set_colorscale(colorScale);
     display_set_graphscale(graphScale);
     display_set_graphrow(jmax/2);
     display_var = dmatrix(mysize, imax);

     mesh_set_plot_var(H, display_var);
     display_set_label("initial values of height");
     display_data(jmax, imax, myoffset, mysize, display_var, x, dx, y, dy);
     sleep(WAIT_TIME);
  }
  
  if(rank==0&&DEBUG>=1) printf("initial values displayed\n");

  if (rank == 0) {
    printf("Iteration:%5.5d, Time:%f, Timestep:%f Total mass:%.10f \n",
        0, time, deltaT, origTM);
  }

  /* run the simulation for given number of iterations */
  gettimeofday(&tstart, NULL);

  for (n = 0; n < ntimes && runflag; n++) {
    mesh_set_boundary_conditions(H, U, V);

    if(rank==0&&DEBUG>=1)printf("Boundary conditions set\n");

    // We calculate a timestep based on SQ(c)=g*h and not let a wave
    // cross a cell.
    gettimeofday(&tstart_sum, NULL);
    deltaT = mesh_set_timestep(g, sigma, H, U, V);
    gettimeofday(&tstop_sum, NULL);
    timersub(&tstop_sum,&tstart_sum,&tresult);
    localtime_reduction += tresult.tv_sec + tresult.tv_usec/1000000.0;

    Cx = deltaT/deltaX;
    Cy = deltaT/deltaY;

    Cxhalf = deltaT/(2.0*deltaX);
    Cyhalf = deltaT/(2.0*deltaY);

    time+=deltaT;
    if(rank==0&&DEBUG>=1){printf("deltaT set to %20.1f\n",deltaT);}

    /* For each element of the matrix ... */
    if (rank == 0&&DEBUG>=1) {printf("Before 1st pass\n");}
    //first pass
    //x direction
    for (j = 0; j < mysize; j++) {
      for (i = 0; i<=imax;i++) {
        //density calculation
        Hx[j][i]=0.5*(H[j+2][i+2]+H[j+2][i+1]) - Cxhalf*( HXFLUX(j,i)-HXFLUX(j,i-1) );
        //momentum x calculation
        Ux[j][i]=0.5*(U[j+2][i+2]+U[j+2][i+1]) - Cxhalf*( UXFLUX(j,i)-UXFLUX(j,i-1) );
        //momentum y calculation
        Vx[j][i]=0.5*(V[j+2][i+2]+V[j+2][i+1]) - Cxhalf*( UVFLUX(j,i)-UVFLUX(j,i-1) );
        if (DEBUG >= 2) {
          printf("1st pass x direction iter %d i %d j %d %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",n,i,j,
              Hx[j][i],Ux[j][i],Vx[j][i],
              H[j+2][i+2],H[j+2][i+1],U[j+2][i+2],U[j+2][i+1],V[j+2][i+2],V[j+2][i+1]);
        }
      }
    }
    if (DEBUG >= 2) {
      printf("\n");
    }
    
    if(rank==0&&DEBUG>=1){printf("First pass x direction complete\n");}
    
    //y direction
    for(j = 0; j<=mysize; j++){
      for(i=0; i<imax; i++){
        //density calculation
        Hy[j][i]=0.5*(H[j+2][i+2]+H[j+1][i+2]) - Cyhalf*( HYFLUX(j,i) - HYFLUX(j-1,i) );
        //momentum x calculation
        Uy[j][i]=0.5*(U[j+2][i+2]+U[j+1][i+2]) - Cyhalf*( UVFLUX(j,i) - UVFLUX(j-1,i) );
        //momentum y calculation
        Vy[j][i]=0.5*(V[j+2][i+2]+V[j+1][i+2]) - Cyhalf*( VYFLUX(j,i) - VYFLUX(j-1,i) );
        if (DEBUG >= 2) {
          printf("1st pass y direction iter %d i %d j %d %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",n,i,j,
              Hy[j][i],Uy[j][i],Vy[j][i],
              H[j+2][i+2],H[j+1][i+2],U[j+2][i+2],U[j+1][i+2],V[j+2][i+2],V[j+1][i+2]);
        }
      }
    }
    if (DEBUG >= 2) {
      printf("\n");
    }

    if(rank==0&&DEBUG>=1){printf("First pass complete\n");}

    //second pass
    if(rank==0&&DEBUG>=1){printf("Second Pass started\n");}
    for (j = 2; j <=mysize+1; j++) {
      for (i = 2; i<=imax+1;i++) {
        duminus1 = H[j][i-1]-H[j][i-2];
        duminus2 = U[j][i-1]-U[j][i-2];
        duplus1 = H[j][i+1]-H[j][i];
        duplus2 = U[j][i+1]-U[j][i];
        duhalf1 = H[j][i]-H[j][i-1];
        duhalf2 = U[j][i]-U[j][i-1];
        rdenom = max(SQ(duhalf1) + SQ(duhalf2),1.0e-30);
        rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
        rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
        rplus =rnumplus /rdenom;
        rminus=rnumminus/rdenom;
        q = max(MIN3(1.0, rminus, rplus), 0.0);
        nu=(fabs(Ux[j-2][i-2])+sqrt(g*Hx[j-2][i-2]))*Cx;
        cv=nu*(1.0-nu);
        wminusx = 0.5*cv*(1.0-q);

        duminus1 = H[j][i]-H[j][i-1];
        duminus2 = U[j][i]-U[j][i-1];
        duplus1 = H[j][i+2]-H[j][i+1];
        duplus2 = U[j][i+2]-U[j][i+1];
        duhalf1 = H[j][i+1]-H[j][i];
        duhalf2 = U[j][i+1]-U[j][i];
        rdenom = max(SQ(duhalf1) + SQ(duhalf2),1.0e-30);
        rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
        rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
        rplus =rnumplus /rdenom;
        rminus=rnumminus/rdenom;
        q = max(MIN3(1.0, rminus, rplus), 0.0);
        nu=(fabs(Ux[j-2][i-1])+sqrt(g*Hx[j-2][i-1]))*Cx;
        cv=nu*(1.0-nu);
        wplusx = 0.5*cv*(1.0-q);

        duminus1 = H[j-1][i]-H[j-2][i];
        duminus2 = V[j-1][i]-V[j-2][i];
        duplus1 = H[j+1][i]-H[j][i];
        duplus2 = V[j+1][i]-V[j][i];
        duhalf1 = H[j][i]-H[j-1][i];
        duhalf2 = V[j][i]-V[j-1][i];
        rdenom = max(SQ(duhalf1) + SQ(duhalf2),1.0e-30);
        rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
        rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
        rplus =rnumplus /rdenom;
        rminus=rnumminus/rdenom;
        q = max(MIN3(1.0, rminus, rplus), 0.0);
        nu=(fabs(Uy[j-2][i-2])+sqrt(g*Hy[j-2][i-2]))*Cy;
        cv=nu*(1.0-nu);
        wminusy = 0.5*cv*(1.0-q);

        duminus1 = H[j][i]-H[j-1][i];
        duminus2 = V[j][i]-V[j-1][i];
        duplus1 = H[j+2][i]-H[j+1][i];
        duplus2 = V[j+2][i]-V[j+1][i];
        duhalf1 = H[j+1][i]-H[j][i];
        duhalf2 = V[j+1][i]-V[j][i];
        rdenom = max(SQ(duhalf1) + SQ(duhalf2),1.0e-30);
        rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
        rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
        rplus =rnumplus /rdenom;
        rminus=rnumminus/rdenom;
        q = max(MIN3(1.0, rminus, rplus), 0.0);
        nu=(fabs(Uy[j-1][i-2])+sqrt(g*Hy[j-1][i-2]))*Cy;
        cv=nu*(1.0-nu);
        wplusy = 0.5*cv*(1.0-q);

        //density calculation
        Hnew[j][i] = H[j][i] - Cx*( HNEWXFLUX(j,i+1) - HNEWXFLUX(j,i) )
                                -wminusx*(H[j][i]-H[j][i-1])+wplusx*(H[j][i+1]-H[j][i])
                             - Cy*( HNEWYFLUX(j+1,i) - HNEWYFLUX(j,i) )
                                -wminusy*(H[j][i]-H[j-1][i])+wplusy*(H[j+1][i]-H[j][i]);
        //momentum x calculation
        Unew[j][i] = U[j][i] - Cx*( UNEWXFLUX(j,i+1) - UNEWXFLUX(j,i) )
                                 -wminusx*(U[j][i]-U[j][i-1])+wplusx*(U[j][i+1]-U[j][i])
                             - Cy*( VUNEWFLUX(j+1,i) - VUNEWFLUX(j,i) )
                                 -wminusy*(U[j][i]-U[j-1][i])+wplusy*(U[j+1][i]-U[j][i]);
        //momentum y calculation
        Vnew[j][i] = V[j][i] - Cx*( UVNEWFLUX(j,i+1) - UVNEWFLUX(j,i) )
                                 -wminusx*(V[j][i]-V[j][i-1])+wplusx*(V[j][i+1]-V[j][i])
                             - Cy*( VNEWYFLUX(j+1,i) - VNEWYFLUX(j,i) )
                                 -wminusy*(V[j][i]-V[j-1][i])+wplusy*(V[j+1][i]-V[j][i]);
        if (DEBUG >= 2) {
          printf("2nd pass iter %d i %d j %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",
              n,i,j,Hnew[j][i],Unew[j][i],Vnew[j][i],
              H[j][i],U[j][i],V[j][i],Hx[j-2][i-2],Hx[j-2][i-1],Ux[j-2][i-2],Ux[j-2][i-1]);
        }
      }
    }
    if (DEBUG >= 2){
      printf("\n");
    }
    if(rank==0&&DEBUG>=1){printf("Second pass complete\n");}
    
    if(rank==0&&DEBUG>=1){printf("Done calculations\n");}

    SWAP(Hnew,H,tmp);
    SWAP(Unew,U,tmp);
    SWAP(Vnew,V,tmp);

    if (DEBUG >= 2) {
      mesh_output("End of Cycle", 0, imax+4, 0, mysize+4, H, U, V);
    }

    gettimeofday(&tstart_sum, NULL);
    TotalMass = sum_var(H);
    gettimeofday(&tstop_sum, NULL);
    timersub(&tstop_sum,&tstart_sum,&tresult);
    localtime_reduction += tresult.tv_sec + tresult.tv_usec/1000000.0;

    if (DISPLAY_ON && (n+1)%DISPLAY_ON==0){
      if (rank == 0) {
         printf("Iteration:%5.5d, Time:%f, Timestep:%f Total mass:%.10f \n",
             n+1, time, deltaT, TotalMass);
      }

      mesh_set_plot_var(H, display_var);
      sprintf(string,"%s     iter %d      time %lf","height",n+1,time);
      display_set_label(string);
      display_data_interactive(jmax, imax, myoffset, mysize, display_var, x, dx, y, dy,
         &ibutton, &key, &xloc, &yloc, &xcor, &ycor, &icell, &jcell, &special_event);

      if (ibutton != -1 || key != '\0'){
         //xycoord_tocell_index(xcor, ycor, icol, jrow);
      }

      //if (ibutton || key !='\0' || special_event || icell >=0 ) {
      //   printf("INPUT -- button is %d key is '%c' special_event is %d icell is %d loc is %d %d coor is %lf %lf\n",
      //         ibutton,      key,        special_event,      icell,      xloc, yloc, xcor, ycor);
      //}
      switch (ibutton){
      case 1:
         break;
      }
      switch(special_event){
      case 1:
         break;
      }
      switch(key){
      case 'E':
         printf("\nKey E: End -- ending calculation\n\n");
         break;
      case 'M':
         printf("        MENU\n");
         printf("\n  Control Keys\n");
         printf("Key E: End     -- ending calculation\n");
         printf("Key M: Menu    -- print out menu for interactive input\n");
         printf("Key Q: Quiet   -- turn off display\n");
         printf("Key V: Verbose -- turn back on display\n");
         printf("\n");
         break;
      case 'Q':
         printf("\nKey Q: Quiet -- turn off display\n\n");
         display_off();
         break;
      case 'V':
         printf("\nKey V: Verbose -- turn back on display\n\n");
         display_on();

         break;
      }

      sleep(WAIT_TIME);
    }
      
    if (CHECK){
      if( fabs(TotalMass-origTM)/origTM>RELATIVE_TOLERANCE || isnan(TotalMass) ){
        printf("Conservation of mass\nMass difference:%e Orig Mass %.10f Current Mass %.10f\n",
            (TotalMass-origTM)/origTM, origTM, TotalMass);
        //printf("Problem occurred on iteration %5.5d at time %f.\n", n+1, time);
        //exit(0);
      }
    }

  }  // End of iteration loop
  
  /* Compute the average time taken/processor */
  gettimeofday(&tstop, NULL);
  timersub(&tstop,&tstart,&tresult);
  localtime = tresult.tv_sec + tresult.tv_usec/1000000.0;
#ifdef USE_MPI
  MPI_Reduce (&localtime, &totaltime, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  elapsed_time = totaltime/(double)numpes;
  elapsed_time_reduction = localtime_reduction;
#else
  elapsed_time = localtime;
  elapsed_time_reduction = localtime_reduction;
#endif
  
  /* Print the total time taken */
  if (rank == 0) {
    printf("[%d] Flow finished in %lf seconds\n",rank,elapsed_time);
    printf("         Reduction in %lf seconds\n",elapsed_time_reduction);
  }

  if(DISPLAY_ON){
    sleep(10);
    display_close();
  }

  // free mesh memory
  memory_free_all();

#ifdef USE_MPI
  sum_finalize();
  MPI_Finalize();
#endif
  exit(0);
}

void mesh_allocation(unsigned int imax, unsigned int mysize)
{
  H = dmatrix(mysize+4, imax+4);
  U = dmatrix(mysize+4, imax+4);
  V = dmatrix(mysize+4, imax+4);

  Hnew = dmatrix(mysize+4, imax+4);
  Unew = dmatrix(mysize+4, imax+4);
  Vnew = dmatrix(mysize+4, imax+4);

  Hx = dmatrix(mysize, imax+1);
  Ux = dmatrix(mysize, imax+1);
  Vx = dmatrix(mysize, imax+1);

  Hy = dmatrix(mysize+1, imax);
  Uy = dmatrix(mysize+1, imax);
  Vy = dmatrix(mysize+1, imax);
}
