/*
 *  ezcl.c
 *
 *  Copyright 2010 Bob Robey, brobey@lanl.gov . All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 *
 */

#ifndef DEBUG
#define DEBUG 0
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/queue.h>
#include <sys/stat.h>

#include "ezcl.h"

// Control flags
struct {
   unsigned int timing     : 1;
} ezcl_flags={1};

cl_uint          nPlatforms;
cl_uint          nDevices;
cl_platform_id  *platforms;
cl_device_id    *devices;
cl_command_queue command_queue;
cl_context       context;
cl_program       program;
cl_platform_id   platform = NULL;

SLIST_HEAD(slist_device_memory_head, device_memory_entry) device_memory_head = SLIST_HEAD_INITIALIZER(device_memory_head);
struct slist_device_memory_head *device_memory_headp;
struct device_memory_entry {
   cl_mem cl_mem_ptr;
   size_t mem_size;
   SLIST_ENTRY(device_memory_entry) device_memory_entries;
} *device_memory_item;

SLIST_HEAD(slist_mapped_memory_head, mapped_memory_entry) mapped_memory_head = SLIST_HEAD_INITIALIZER(mapped_memory_head);
struct slist_mapped_memory_head *mapped_memory_headp;
struct mapped_memory_entry {
   cl_mem cl_mem_ptr;
   size_t mem_size;
   SLIST_ENTRY(mapped_memory_entry) mapped_memory_entries;
} *mapped_memory_item;

SLIST_HEAD(slist_malloc_memory_head, malloc_memory_entry) malloc_memory_head = SLIST_HEAD_INITIALIZER(malloc_memory_head);
struct slist_malloc_memory_head *malloc_memory_headp;
struct malloc_memory_entry {
   void *mem_ptr;
   size_t mem_size;
   SLIST_ENTRY(malloc_memory_entry) malloc_memory_entries;
} *malloc_memory_item;

SLIST_HEAD(slist_object_head, object_entry) object_head = SLIST_HEAD_INITIALIZER(object_head);
struct slist_object_head *object_headp;
struct object_entry {
   int object_type;
   cl_program program;
   cl_kernel kernel;
   cl_command_queue command_queue;
   cl_context context;
   char *filename;
   SLIST_ENTRY(object_entry) object_entries;
} *object_item;


cl_int ezcl_init(cl_context *ezcl_gpu_context, cl_context *ezcl_cpu_context, cl_context *ezcl_accelerator_context)
{
   int ierr;
   cl_uint nPlatforms;
   cl_uint nDevices = 0;

   ierr = clGetPlatformIDs(0, 0, &nPlatforms);
   if (ierr != CL_SUCCESS){
      printf("EZCL_INIT: Error with clGetDeviceIDs call at line %d in file %s\n",__LINE__,__FILE__);
      if (ierr == CL_INVALID_VALUE){
         printf("Invalid value in clGetPlatformID call\n");
      }
      exit(-1);
   }
   if (nPlatforms == 0) {
      printf("EZCL_INIT: Error -- No opencl platforms detected at line %d in file %s\n",__LINE__,__FILE__);
      exit(-1);
   }
   if (DEBUG){
      printf("\n\nEZCL_INIT: %d opencl platform(s) detected\n",nPlatforms);
   }
   
   cl_uint num_platforms_requested = nPlatforms;
   platforms = (cl_platform_id *)malloc(nPlatforms*sizeof(cl_platform_id));
   ezcl_malloc_memory_add(platforms, nPlatforms*sizeof(cl_platform_id));
   
   ierr = clGetPlatformIDs(num_platforms_requested, platforms, &nPlatforms);
   if (ierr != CL_SUCCESS || nPlatforms != num_platforms_requested){
      printf("EZCL_INIT: Error with clGetPlatformIDs call at line %d in file %s\n",__LINE__,__FILE__);
        if (ierr == CL_INVALID_VALUE){
          printf("Invalid value in clGetPlatformID call\n");
        }
   }

   if (DEBUG){
      char info[1024];
      for (int iplatform=0; iplatform<nPlatforms; iplatform++){
         printf("  Platform %d:\n",iplatform+1);
      
         //clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_PROFILE,   1024,info,0);
         //printf("    CL_PLATFORM_PROFILE    : %s\n",info);
      
         clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_VERSION,   1024,info,0);
         printf("    CL_PLATFORM_VERSION    : %s\n",info);
      
         clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_NAME,      1024,info,0);
         printf("    CL_PLATFORM_NAME       : %s\n",info);
      
         clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_VENDOR,    1024,info,0);
         printf("    CL_PLATFORM_VENDOR     : %s\n",info);
      
         //clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_EXTENSIONS,1024,info,0);
         //printf("    CL_PLATFORM_EXTENSIONS : %s\n",info);
      }
      printf("\n");
   }
   
   clGetDeviceIDs(platforms[0],CL_DEVICE_TYPE_ALL,0,0,&nDevices);
   if (nDevices == 0) {
      printf("EZCL_INIT: Error -- No opencl devices detected at line %d in file %s\n",__LINE__,__FILE__);
      //exit(-1);
   }
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PLATFORM:
      *  CL_INVALID_DEVICE_TYPE:
      *  CL_INVALID_VALUE:
      *  CL_DEVICE_NOT_FOUND:
      */
     ezcl_print_error(ierr, "EZCL_INIT", "clGetDeviceIDs", __LINE__, __FILE__);
   }
   if (DEBUG){
      printf("EZCL_INIT: %d opencl devices(s) detected at line\n",nDevices);
   }
   
   cl_uint num_devices_requested = nDevices;
   devices = (cl_device_id *)malloc(nDevices*sizeof(cl_device_id));
   ezcl_malloc_memory_add(devices, nDevices*sizeof(cl_device_id));

   ierr = clGetDeviceIDs(platforms[0],CL_DEVICE_TYPE_ALL,num_devices_requested,devices,&nDevices);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PLATFORM:
      *  CL_INVALID_DEVICE_TYPE:
      *  CL_INVALID_VALUE:
      *  CL_DEVICE_NOT_FOUND:
      */
     ezcl_print_error(ierr, "EZCL_INIT", "clGetDeviceIDs", __LINE__, __FILE__);
   }
   
   if (DEBUG){
      for (int idevice=0; idevice<nDevices; idevice++){
         printf(  "  Device %d:\n", idevice+1);
         ezcl_device_info(devices[idevice]);
      }
   }
   
   *ezcl_gpu_context = clCreateContextFromType(NULL, CL_DEVICE_TYPE_GPU, NULL, NULL, &ierr);
   if (ierr == CL_INVALID_VALUE){
      printf("Invalid value in clCreateContext call\n");
      if (devices == NULL) printf("Devices is NULL\n");
   }
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PLATFORM:
      *  CL_INVALID_VALUE:
      *  CL_INVALID_DEVICE_TYPE:
      *  CL_DEVICE_NOT_AVAILABLE:
      *  CL_DEVICE_NOT_FOUND:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_INIT", "clCreateContextFromType", __LINE__, __FILE__);
   }
   
   if (*ezcl_gpu_context != NULL){
      printf("EZCL_INIT: GPU device context created\n");
   } else {
      if (DEBUG) printf("EZCL_INIT: No gpu device found at line %d in file %s\n",__LINE__,__FILE__);
   }
   
   *ezcl_cpu_context = clCreateContextFromType(0, CL_DEVICE_TYPE_CPU, NULL, NULL, &ierr);
   if (ierr == CL_INVALID_VALUE){
      printf("Invalid value in clCreateContext call\n");
      if (devices == NULL) printf("Devices is NULL\n");
   }
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PLATFORM:
      *  CL_INVALID_VALUE:
      *  CL_INVALID_DEVICE_TYPE:
      *  CL_DEVICE_NOT_AVAILABLE:
      *  CL_DEVICE_NOT_FOUND:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_INIT", "clCreateContextFromType", __LINE__, __FILE__);
   }
   if (*ezcl_cpu_context != NULL){
      printf("EZCL_INIT: CPU device context created\n");
   } else {
      if (DEBUG) printf("EZCL_INIT: No cpu device found\n");
   }

   *ezcl_accelerator_context = clCreateContextFromType(0, CL_DEVICE_TYPE_ACCELERATOR, NULL, NULL, &ierr);
   if (ierr == CL_INVALID_VALUE){
      printf("Invalid value in clCreateContext call\n");
      if (devices == NULL) printf("Devices is NULL\n");
   }
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PLATFORM:
      *  CL_INVALID_VALUE:
      *  CL_INVALID_DEVICE_TYPE:
      *  CL_DEVICE_NOT_AVAILABLE:
      *  CL_DEVICE_NOT_FOUND:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_INIT", "clCreateContextFromType", __LINE__, __FILE__);
   }
   if (*ezcl_accelerator_context != NULL){
      printf("EZCL_INIT: Accelerator device context created\n");
   } else {
      if (DEBUG) printf("EZCL_INIT: No accelerator device found at line %d in file %s\n",__LINE__,__FILE__);
   }
    printf("\n");
   
   return(0);
}

cl_int ezcl_devtype_init(cl_device_type device_type, cl_context *return_context, cl_command_queue *return_command_queue){
   cl_device_id device;
   int ierr;

   cl_uint nPlatforms;
   cl_uint nDevices;
   cl_uint platform_selected = -1;

   // Get the number of platforms first, then allocate and get the platform
   ierr = clGetPlatformIDs(0, NULL, &nPlatforms);
   if (ierr != CL_SUCCESS){
      printf("EZCL_DEVTYPE_INIT: Error with clGetDeviceIDs call at line %d in file %s\n",__LINE__,__FILE__);
      if (ierr == CL_INVALID_VALUE){
         printf("Invalid value in clGetPlatformID call\n");
      }
      exit(-1);
   }
   if (nPlatforms == 0) {
      printf("EZCL_DEVTYPE_INIT: Error -- No opencl platforms detected at line %d in file %s\n",__LINE__,__FILE__);
      exit(-1);
   }
   if (DEBUG){
      printf("\n\nEZCL_DEVTYPE_INIT: %d opencl platform(s) detected\n",nPlatforms);
   }

   platforms = (cl_platform_id *)malloc(nPlatforms*sizeof(cl_platform_id));
   ezcl_malloc_memory_add(platforms, nPlatforms*sizeof(cl_platform_id));

   ierr = clGetPlatformIDs(nPlatforms, platforms, NULL);
   if (ierr != CL_SUCCESS){
      printf("EZCL_DEVTYPE_INIT: Error with clGetPlatformIDs call at line %d in file %s\n",__LINE__,__FILE__);
      if (ierr == CL_INVALID_VALUE){
         printf("Invalid value in clGetPlatformID call\n");
      }
   }

   if (DEBUG){
      char info[1024];
      for (int iplatform=0; iplatform<nPlatforms; iplatform++){
         printf("  Platform %d:\n",iplatform+1);

         //clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_PROFILE,   1024,info,0);
         //printf("    CL_PLATFORM_PROFILE    : %s\n",info);

         clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_VERSION,   1024,info,0);
         printf("    CL_PLATFORM_VERSION    : %s\n",info);

         clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_NAME,      1024,info,0);
         printf("    CL_PLATFORM_NAME       : %s\n",info);

         clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_VENDOR,    1024,info,0);
         printf("    CL_PLATFORM_VENDOR     : %s\n",info);

         //clGetPlatformInfo(platforms[iplatform],CL_PLATFORM_EXTENSIONS,1024,info,0);
         //printf("    CL_PLATFORM_EXTENSIONS : %s\n",info);
      }
      printf("\n");
   }

   // Get the number of devices, allocate, and get the devices
   for (int iplatform=0; iplatform<nPlatforms; iplatform++){
      ierr = clGetDeviceIDs(platforms[0],device_type,0,NULL,&nDevices);
      if (ierr == CL_DEVICE_NOT_FOUND) {
         if (DEBUG) {
           printf("Warning: Device of requested type not found for platform %d in clGetDeviceID call\n",iplatform);
         }
         continue;
      }
      if (ierr != CL_SUCCESS) {
        /* Possible Errors
         *  CL_INVALID_PLATFORM:
         *  CL_INVALID_DEVICE_TYPE:
         *  CL_INVALID_VALUE:
         *  CL_DEVICE_NOT_FOUND:
         */
        ezcl_print_error(ierr, "EZCL_DEVTYPE_INIT", "clGetDeviceIDs", __LINE__, __FILE__);
      }
      if (DEBUG){
         printf("EZCL_DEVTYPE_INIT: %d opencl devices(s) detected\n",nDevices);
      }
      platform_selected = iplatform;
      platform = platforms[iplatform];
      exit;
   }

   if (platform_selected == -1){
      printf("Warning: Device of requested type not found in clGetDeviceID call\n");
      return(EZCL_NODEVICE);
   }

   devices = (cl_device_id *)malloc(nDevices*sizeof(cl_device_id));
   ezcl_malloc_memory_add(devices, nDevices*sizeof(cl_device_id));

   ierr = clGetDeviceIDs(platforms[platform_selected],device_type,nDevices,devices,NULL);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PLATFORM:
      *  CL_INVALID_DEVICE_TYPE:
      *  CL_INVALID_VALUE:
      *  CL_DEVICE_NOT_FOUND:
      */
     ezcl_print_error(ierr, "EZCL_DEVTYPE_INIT", "clGetDeviceIDs", __LINE__, __FILE__);
   }

   if (DEBUG){
      for (int idevice=0; idevice<nDevices; idevice++){
         printf(  "  Device %d:\n", idevice+1);
         ezcl_device_info(devices[idevice]);
      }
   }

   cl_context_properties context_properties[3]=
   {
     CL_CONTEXT_PLATFORM,
     (cl_context_properties)platform,
     0
   };

   context = clCreateContext(context_properties, 1, devices, NULL, NULL, &ierr);
   if (ierr == CL_INVALID_VALUE){
      printf("Invalid value in clCreateContext call\n");
      if (devices == NULL) printf("Devices is NULL\n");
   }
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PLATFORM:
      *  CL_INVALID_VALUE:
      *  CL_INVALID_DEVICE_TYPE:
      *  CL_DEVICE_NOT_AVAILABLE:
      *  CL_DEVICE_NOT_FOUND:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_DEVTYPE_INIT", "clCreateContext", __LINE__, __FILE__);
   }
   if (DEBUG){
      if (context != NULL){
         if(device_type & CL_DEVICE_TYPE_CPU )
            printf("EZCL_DEVTYPE_INIT: CPU device context created\n");
         else if (device_type & CL_DEVICE_TYPE_GPU){
            printf("EZCL_DEVTYPE_INIT: GPU device context created\n");
         }
         else if (device_type & CL_DEVICE_TYPE_ACCELERATOR){
            printf("EZCL_DEVTYPE_INIT: ACCELERATOR device context created\n");
         }
         else if (device_type & CL_DEVICE_TYPE_DEFAULT){
            printf("EZCL_DEVTYPE_INIT: Default device context created\n");
         }
      } else {
         if (DEBUG == 2) printf("EZCL_DEVTYPE_INIT: No device of type specified found\n");
      }
   }

   *return_context = ezcl_get_context();

   object_item = malloc(sizeof(struct object_entry));
   object_item->object_type = CONTEXT_OBJECT;
   object_item->context = context;
   if (DEBUG) printf("EZCL_DEVTYPE_INIT: DEBUG -- context is %p\n",context);
   SLIST_INSERT_HEAD(&object_head, object_item, object_entries);

   clGetContextInfo(context, CL_CONTEXT_DEVICES, sizeof(device), &device, NULL);
   if (context == NULL){
      printf("EZCL_DEVTYPE_INIT: Failed to find device and setup context at line %d in file %s\n",__LINE__,__FILE__);
      exit(-1); /* No device is available, something is wrong */
   }
   if (DEBUG == 2){
      ezcl_device_info(device);
   }

   command_queue = ezcl_create_command_queue(context);
   *return_command_queue = ezcl_get_command_queue();
   
   return(EZCL_SUCCESS);
}

cl_device_id ezcl_get_device(cl_context context){
   cl_device_id device;
   int ierr;
   ierr = clGetContextInfo(context, CL_CONTEXT_DEVICES, sizeof(device), &device, NULL);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_CONTEXT:
      *  CL_INVALID_VALUE:
      */
     ezcl_print_error(ierr, "EZCL_DEVTYPE_INIT", "clGetContextInfo", __LINE__, __FILE__);
   }
   return(device);
}

void ezcl_device_info(cl_device_id device){
   char info[1024];
   cl_bool iflag;
   cl_uint inum;
   size_t isize;
   cl_ulong ilong;
   cl_device_type device_type;
   cl_command_queue_properties iprop;

   clGetDeviceInfo(device,CL_DEVICE_TYPE,sizeof(device_type),&device_type,0);
   if( device_type & CL_DEVICE_TYPE_CPU )
      printf("    CL_DEVICE_TYPE                       : %s\n", "CL_DEVICE_TYPE_CPU");
   if( device_type & CL_DEVICE_TYPE_GPU )
      printf("    CL_DEVICE_TYPE                       : %s\n", "CL_DEVICE_TYPE_GPU");
   if( device_type & CL_DEVICE_TYPE_ACCELERATOR )
      printf("    CL_DEVICE_TYPE                       : %s\n", "CL_DEVICE_TYPE_ACCELERATOR");
   if( device_type & CL_DEVICE_TYPE_DEFAULT )
      printf("    CL_DEVICE_TYPE                       : %s\n", "CL_DEVICE_TYPE_DEFAULT");
               
   clGetDeviceInfo(device,CL_DEVICE_AVAILABLE,sizeof(iflag),&iflag,0);
   if (iflag == CL_TRUE) {
      printf(  "    CL_DEVICE_AVAILABLE                  : TRUE\n");
   } else {
      printf(  "    CL_DEVICE_AVAILABLE                  : FALSE\n");
   }
   
   clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(info), &info, NULL);
   printf(  "    CL_DEVICE_VENDOR                     : %s\n", info);
   
   clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(info), &info, NULL);
   printf(  "    CL_DEVICE_NAME                       : %s\n", info);
   
   clGetDeviceInfo(device, CL_DRIVER_VERSION, sizeof(info), &info, NULL);
   printf(  "    CL_DRIVER_VERSION                    : %s\n", info);
   
   clGetDeviceInfo(device, CL_DEVICE_VERSION, sizeof(info), &info, NULL);
   printf(  "    CL_DEVICE_VERSION                    : %s\n", info);
   
   clGetDeviceInfo(device,CL_DEVICE_MAX_COMPUTE_UNITS,sizeof(inum),&inum,0);
   printf(  "    CL_DEVICE_MAX_COMPUTE_UNITS          : %d\n", inum);
   
   clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,sizeof(inum),&inum,0);
   printf(  "    CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS   : %d\n", inum);
   
   size_t *item_sizes = (size_t *)malloc(inum*sizeof(size_t));
   clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_SIZES,sizeof(item_sizes),item_sizes,0);
   printf(  "    CL_DEVICE_MAX_WORK_ITEM_SIZES        : %ld %ld %ld\n",
         item_sizes[0], item_sizes[1], item_sizes[2]);
   free(item_sizes);
   
   clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_GROUP_SIZE,sizeof(isize),&isize,0);
   printf(  "    CL_DEVICE_MAX_WORK_GROUP_SIZE        : %ld\n", isize);
   
   clGetDeviceInfo(device,CL_DEVICE_MAX_CLOCK_FREQUENCY,sizeof(inum),&inum,0);
   printf(  "    CL_DEVICE_MAX_CLOCK_FREQUENCY        : %d\n", inum);
   
   clGetDeviceInfo(device,CL_DEVICE_MAX_MEM_ALLOC_SIZE,sizeof(inum),&inum,0);
   printf(  "    CL_DEVICE_MAX_MEM_ALLOC_SIZE         : %d\n", inum);
   
#ifdef __APPLE_CC__
   clGetDeviceInfo(device,CL_DEVICE_GLOBAL_MEM_SIZE,sizeof(ilong),&ilong,0);
   printf(  "    CL_DEVICE_GLOBAL_MEM_SIZE            : %llu\n", ilong);
   
   clGetDeviceInfo(device,CL_DEVICE_GLOBAL_MEM_CACHE_SIZE,sizeof(ilong),&ilong,0);
   printf(  "    CL_DEVICE_GLOBAL_MEM_CACHE_SIZE      : %llu\n", ilong);
#else
   clGetDeviceInfo(device,CL_DEVICE_GLOBAL_MEM_SIZE,sizeof(ilong),&ilong,0);
   printf(  "    CL_DEVICE_GLOBAL_MEM_SIZE            : %lu\n", ilong);

   clGetDeviceInfo(device,CL_DEVICE_GLOBAL_MEM_CACHE_SIZE,sizeof(ilong),&ilong,0);
   printf(  "    CL_DEVICE_GLOBAL_MEM_CACHE_SIZE      : %lu\n", ilong);
#endif
   clGetDeviceInfo(device,CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE,sizeof(inum),&inum,0);
   printf(  "    CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE  : %d\n", inum);
   
   clGetDeviceInfo(device,CL_DEVICE_MAX_CONSTANT_ARGS,sizeof(inum),&inum,0);
   printf(  "    CL_DEVICE_GLOBAL_MAX_CONSTANT_ARGS   : %d\n", inum);
   
   clGetDeviceInfo(device,CL_DEVICE_ERROR_CORRECTION_SUPPORT,sizeof(iflag),&iflag,0);
   if (iflag == CL_TRUE) {
      printf(  "    CL_DEVICE_ERROR_CORRECTION_SUPPORT   : TRUE\n");
   } else {
      printf(  "    CL_DEVICE_ERROR_CORRECTION_SUPPORT   : FALSE\n");
   }
   
   clGetDeviceInfo(device,CL_DEVICE_PROFILING_TIMER_RESOLUTION,sizeof(isize),&isize,0);
   printf(  "    CL_DEVICE_PROFILING_TIMER_RESOLUTION : %ld nanosecs\n", isize);
   
   clGetDeviceInfo(device,CL_DEVICE_QUEUE_PROPERTIES,sizeof(iprop),&iprop,0);
   if (iprop & CL_QUEUE_PROFILING_ENABLE) {
      printf(  "    CL_DEVICE_QUEUE PROFILING            : AVAILABLE\n");
   } else {
      printf(  "    CL_DEVICE_QUEUE PROFILING            : NOT AVAILABLE\n");
   }
   
   clGetDeviceInfo(device, CL_DEVICE_EXTENSIONS, sizeof(info), &info, NULL);
   printf(  "    CL_DEVICE_EXTENSIONS                 : %s\n\n", info);
   
}

cl_command_queue ezcl_create_command_queue(cl_context context){
   cl_device_id device;
   int ierr;
   cl_command_queue_properties queueProps;
   ierr = clGetContextInfo(context, CL_CONTEXT_DEVICES, sizeof(device), &device, NULL);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_CONTEXT:
      *  CL_INVALID_VALUE:
      */
     ezcl_print_error(ierr, "EZCL_CREATE_COMMAND_QUEUE", "clGetContextInfo", __LINE__, __FILE__);
   }
   queueProps = ezcl_flags.timing ? CL_QUEUE_PROFILING_ENABLE : 0;
   command_queue = clCreateCommandQueue(context, device, queueProps, &ierr);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_CONTEXT:
      *  CL_INVALID_DEVICE:
      *  CL_INVALID_VALUE:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_CREATE_COMMAND_QUEUE", "clCreateCommandQueue", __LINE__, __FILE__);
   }

   object_item = malloc(sizeof(struct object_entry));
   object_item->object_type = COMMAND_QUEUE_OBJECT;
   object_item->command_queue = command_queue;
   if (DEBUG) printf("EZCL_DEVTYPE_INIT: DEBUG -- command_queue is %p\n",command_queue);
   SLIST_INSERT_HEAD(&object_head, object_item, object_entries);

   return(command_queue);
}

cl_command_queue ezcl_get_command_queue(void){
   return(command_queue);
}

cl_context ezcl_get_context(void){
   return(context);
}

cl_mem ezcl_malloc(void *host_mem_ptr, size_t dims[], size_t elsize, int flags, int ezcl_flags){
   cl_int ierr;
   void *buf_mem_ptr = NULL;
   cl_mem mem_ptr = NULL;
   size_t size=0;

   size=dims[0]*elsize;
   //printf("size is %d\n",size);

   if (flags & CL_MEM_ALLOC_HOST_PTR){
      if (ezcl_flags & EZCL_PINNED_MEMORY){
         buf_mem_ptr = clCreateBuffer(context, flags, size, host_mem_ptr, &ierr);
         if (ierr != CL_SUCCESS) {
           /* Possible Errors
            *  CL_INVALID_CONTEXT:
            *  CL_INVALID_VALUE:
            *  CL_INVALID_BUFFER_SIZE:
            *  CL_INVALID_HOST_PTR:
            *  CL_MEM_OBJECT_ALLOCATION_FAILURE:
            *  CL_OUT_OF_HOST_MEMORY:
            */
           ezcl_print_error(ierr, "EZCL_MALLOC", "clCreateBuffer", __LINE__, __FILE__);
         }
         ezcl_device_memory_add(buf_mem_ptr, size);

         if (flags & CL_MEM_READ_ONLY){
           mem_ptr = clEnqueueMapBuffer(command_queue, buf_mem_ptr, CL_TRUE, CL_MAP_WRITE, 0, size, 0, NULL, NULL, NULL);
         }else if (flags & CL_MEM_WRITE_ONLY){
           mem_ptr = clEnqueueMapBuffer(command_queue, buf_mem_ptr, CL_TRUE, CL_MAP_READ, 0, size, 0, NULL, NULL, NULL);
         } else {
           mem_ptr = clEnqueueMapBuffer(command_queue, buf_mem_ptr, CL_TRUE, 0, 0, size, 0, NULL, NULL, NULL);
         }
         ezcl_mapped_memory_add(mem_ptr, size);

      } else {
         mem_ptr = malloc(size);
         ezcl_malloc_memory_add(mem_ptr, size);
      }
   } else {
      mem_ptr = clCreateBuffer(context, flags, size, host_mem_ptr, &ierr);
      if (ierr != CL_SUCCESS) {
        /* Possible Errors
         *  CL_INVALID_CONTEXT:
         *  CL_INVALID_VALUE:
         *  CL_INVALID_BUFFER_SIZE:
         *  CL_INVALID_HOST_PTR:
         *  CL_MEM_OBJECT_ALLOCATION_FAILURE:
         *  CL_OUT_OF_HOST_MEMORY:
         */
        ezcl_print_error(ierr, "EZCL_MALLOC", "clCreateBuffer", __LINE__, __FILE__);
      }
      ezcl_device_memory_add(mem_ptr, size);
   }

   return(mem_ptr);
}

void ezcl_device_memory_add(cl_mem dev_mem_ptr, size_t size){
   if (SLIST_EMPTY(&device_memory_head)) SLIST_INIT(&device_memory_head);

   device_memory_item = malloc(sizeof(struct device_memory_entry));
   device_memory_item->cl_mem_ptr = dev_mem_ptr;
   device_memory_item->mem_size=size;
   if (DEBUG) printf("EZCL_DEVICE_MEMORY_ADD: DEBUG -- cl memory pointer is %p\n",dev_mem_ptr);
   SLIST_INSERT_HEAD(&device_memory_head, device_memory_item, device_memory_entries);
}

void ezcl_mapped_memory_add(cl_mem map_mem_ptr, size_t size){
   if (SLIST_EMPTY(&mapped_memory_head)) SLIST_INIT(&mapped_memory_head);

   mapped_memory_item = malloc(sizeof(struct mapped_memory_entry));
   mapped_memory_item->cl_mem_ptr = map_mem_ptr;
   mapped_memory_item->mem_size=size;
   if (DEBUG) printf("EZCL_MAPPED_MEMORY_ADD: DEBUG -- cl memory pointer is %p\n",map_mem_ptr);
   SLIST_INSERT_HEAD(&mapped_memory_head, mapped_memory_item, mapped_memory_entries);
}

void *ezcl_malloc_memory_add(void *malloc_mem_ptr, size_t size){
   if (SLIST_EMPTY(&malloc_memory_head)) SLIST_INIT(&malloc_memory_head);
   
   malloc_memory_item = malloc(sizeof(struct malloc_memory_entry));
   malloc_memory_item->mem_ptr = malloc_mem_ptr;
   malloc_memory_item->mem_size = size;
   if (DEBUG) printf("EZCL_MALLOC_MEMORY_ADD: DEBUG -- malloc memory pointer is %p\n",malloc_mem_ptr);

   SLIST_INSERT_HEAD(&malloc_memory_head, malloc_memory_item, malloc_memory_entries);

   return(malloc_mem_ptr);
}

void ezcl_device_memory_remove(void *dev_mem_ptr){
   SLIST_FOREACH(device_memory_item, &device_memory_head, device_memory_entries){
      if (device_memory_item->cl_mem_ptr == dev_mem_ptr) {
         if (DEBUG) printf("EZCL_DEVICE_MEMORY_REMOVE: DEBUG -- freeing device memory pointer %p\n",dev_mem_ptr);
         clReleaseMemObject(device_memory_item->cl_mem_ptr);
         SLIST_REMOVE(&device_memory_head, device_memory_item, device_memory_entry, device_memory_entries);
         free(device_memory_item);
            break;         
      }
   }
}   

void ezcl_mapped_memory_remove(void *map_mem_ptr){
   SLIST_FOREACH(mapped_memory_item, &mapped_memory_head, mapped_memory_entries){
      if (mapped_memory_item->cl_mem_ptr == map_mem_ptr) {
         if (DEBUG) printf("EZCL_MAPPED_MEMORY_REMOVE: DEBUG -- freeing mapped memory pointer %p\n",map_mem_ptr);
         clReleaseMemObject(mapped_memory_item->cl_mem_ptr);
         SLIST_REMOVE(&mapped_memory_head, mapped_memory_item, mapped_memory_entry, mapped_memory_entries);
         free(mapped_memory_item);
         break;
      }
   }
}

void ezcl_malloc_memory_remove(void *malloc_mem_ptr){
   SLIST_FOREACH(malloc_memory_item, &malloc_memory_head, malloc_memory_entries){
      if (malloc_memory_item->mem_ptr == malloc_mem_ptr) {
         if (DEBUG) printf("EZCL_MALLOC_MEMORY_REMOVE: DEBUG -- freeing malloc memory pointer %p\n",malloc_mem_ptr);
         free(malloc_mem_ptr);
         SLIST_REMOVE(&malloc_memory_head, malloc_memory_item, malloc_memory_entry, malloc_memory_entries);
         free(malloc_memory_item);
         break;         
      }
   }
}   

void ezcl_program_release(cl_program program){
    SLIST_FOREACH(object_item, &object_head, object_entries){
        if (object_item->object_type == PROGRAM_OBJECT && object_item->program == program) {
            if (DEBUG) printf("EZCL_PROGRAM_RELEASE: DEBUG -- releasing program %p\n",program);
            clReleaseProgram(program);
            SLIST_REMOVE(&object_head, object_item, object_entry, object_entries);
            free(object_item);
            break;
        }
    }
}

void ezcl_kernel_release(cl_kernel kernel){
    SLIST_FOREACH(object_item, &object_head, object_entries){
        if (object_item->object_type == KERNEL_OBJECT && object_item->kernel == kernel) {
            if (DEBUG) printf("EZCL_KERNEL_RELEASE: DEBUG -- releasing kernel %p\n",kernel);
            clReleaseKernel(kernel);
            SLIST_REMOVE(&object_head, object_item, object_entry, object_entries);
            free(object_item);
            break;
        }
    }
}

void ezcl_command_queue_release(cl_command_queue command_queue){
    SLIST_FOREACH(object_item, &object_head, object_entries){
        if (object_item->object_type == COMMAND_QUEUE_OBJECT && object_item->command_queue == command_queue) {
            if (DEBUG) printf("EZCL_COMMAND_QUEUE_RELEASE: DEBUG -- releasing command queue %p\n",command_queue);
            clReleaseCommandQueue(command_queue);
            SLIST_REMOVE(&object_head, object_item, object_entry, object_entries);
            free(object_item);
            break;
        }
    }
}

void ezcl_context_release(cl_context context){
    SLIST_FOREACH(object_item, &object_head, object_entries){
        if (object_item->object_type == CONTEXT_OBJECT && object_item->context == context) {
            if (DEBUG) printf("EZCL_CONTEXT_RELEASE: DEBUG -- releasing context %p\n",context);
            clReleaseContext(context);
            SLIST_REMOVE(&object_head, object_item, object_entry, object_entries);
            free(object_item);
            break;
        }
    }
}

void ezcl_mem_free_all(void){
   while (!SLIST_EMPTY(&device_memory_head)) {
      device_memory_item = SLIST_FIRST(&device_memory_head);
      if (DEBUG) printf("EZCL_MEM_FREE_ALL: DEBUG -- freeing cl memory %p\n",device_memory_item->cl_mem_ptr);
      clReleaseMemObject((cl_mem)device_memory_item->cl_mem_ptr);
      SLIST_REMOVE_HEAD(&device_memory_head, device_memory_entries);
      free(device_memory_item);
   }

   while (!SLIST_EMPTY(&malloc_memory_head)) {
      malloc_memory_item = SLIST_FIRST(&malloc_memory_head);
      if (DEBUG) printf("EZCL_MEM_FREE_ALL: DEBUG -- freeing malloc memory %p\n",malloc_memory_item->mem_ptr);
      free(malloc_memory_item->mem_ptr);
      SLIST_REMOVE_HEAD(&malloc_memory_head, malloc_memory_entries);
      free(malloc_memory_item);
   }
   
   while (!SLIST_EMPTY(&object_head)) {
      object_item = SLIST_FIRST(&object_head);
      switch(object_item->object_type){
      case PROGRAM_OBJECT:
        if (DEBUG) printf("EZCL_MEM_FREE_ALL: DEBUG -- releasing program %p %s\n",object_item->program, object_item->filename);
        free(object_item->filename);
        clReleaseProgram(object_item->program);
        break;
        case KERNEL_OBJECT:
          if (DEBUG) printf("EZCL_MEM_FREE_ALL: DEBUG -- releasing kernel %p\n",object_item->kernel);
          clReleaseKernel(object_item->kernel);
          break;
        case COMMAND_QUEUE_OBJECT:
          if (DEBUG) printf("EZCL_MEM_FREE_ALL: DEBUG -- releasing command_queue %p\n",object_item->command_queue);
          clReleaseCommandQueue(object_item->command_queue);
          break;
        case CONTEXT_OBJECT:
          if (DEBUG) printf("EZCL_MEM_FREE_ALL: DEBUG -- releasing context %p\n",object_item->context);
          clReleaseContext(object_item->context);
          break;
      }
      SLIST_REMOVE_HEAD(&object_head, object_entries);
      free(object_item);
   }
   
}

cl_kernel ezcl_create_kernel(cl_context context, const char *filename, const char *kernel_name, int flags){
   cl_kernel kernel;
   struct stat statbuf;
   FILE *fh;
   char *source;
   cl_int ierr;
   size_t nReportSize;
   char *BuildReport;
   char *filename_copy;
   
   fh=fopen(filename, "r");
   if (fh == 0){
      printf("EZCL_CREATE_KERNEL: Error -- cannot open %s at line %d in file %s\n",filename,__LINE__,__FILE__);
      exit(-1);
   }
   
   stat(filename, &statbuf);
   source = (char *)malloc(statbuf.st_size + 1);
   ezcl_malloc_memory_add(source, statbuf.st_size + 1);
   
   if (fread(source, statbuf.st_size, 1, fh) != 1) {
      printf("ERROR: problem reading program source file %s\n",filename);
      exit(-1);
   }
   source[statbuf.st_size] = '\0';

   program = clCreateProgramWithSource(context, 1, (const char **)&source, NULL, &ierr);
   if (ierr != CL_SUCCESS){
      printf("EZCL_CREATE_KERNEL: clBuildProgramWithSource returned an error %d at line %d inf file %s\n",ierr,__LINE__,__FILE__);
      switch (ierr){
      case CL_INVALID_CONTEXT:
       printf("Invalid context in clBuildProgramWithSource\n");
       break;
      case CL_INVALID_VALUE:
        printf("Invalid value in clBuildProgramWithSource\n");
        break;
      case CL_OUT_OF_HOST_MEMORY:
        printf("Out of host memory in clBuildProgramWithSource\n");
        break;
     }
   }
   
#ifdef HAVE_CL_DOUBLE
   ierr = clBuildProgram(program, 0, NULL, "-DHAVE_CL_DOUBLE", NULL, NULL);
#else
   ierr = clBuildProgram(program, 0, NULL, "-DNO_CL_DOUBLE", NULL, NULL);
#endif
   if (ierr != CL_SUCCESS){
      printf("EZCL_CREATE_KERNEL: clBuildProgram returned an error %d at line %d in file %s\n",ierr,__LINE__,__FILE__);
        switch (ierr){
        case CL_INVALID_PROGRAM:
          printf("Invalid program in clBuildProgram\n");
          break;
        case CL_INVALID_VALUE:
          printf("Invalid value in clBuildProgram\n");
          break;
        case CL_INVALID_DEVICE:
          printf("Invalid device in clBuildProgram\n");
          break;
        case CL_INVALID_BUILD_OPTIONS:
          printf("Invalid build options in clBuildProgram\n");
          break;
        case CL_INVALID_OPERATION:
          printf("Invalid operation in clBuildProgram\n");
          break;
        case CL_COMPILER_NOT_AVAILABLE:
          printf("CL compiler not available in clBuildProgram\n");
          break;
        case CL_BUILD_PROGRAM_FAILURE:
          printf("Build program failure in clBuildProgram\n");
          ierr = clGetProgramBuildInfo(program, ezcl_get_device(context), CL_PROGRAM_BUILD_LOG, 0, NULL, &nReportSize);
          if (ierr != CL_SUCCESS) {
            switch (ierr){
               case CL_INVALID_DEVICE:
                  printf("Invalid device in clProgramBuildInfo\n");
                  break;
               case CL_INVALID_VALUE:
                  printf("Invalid value in clProgramBuildInfo\n");
                  break;
               case CL_INVALID_PROGRAM:
                  printf("Invalid program in clProgramBuildInfo\n");
                  break;
               }
            }
                 
            BuildReport = (char *)malloc(nReportSize);
                 
            ierr = clGetProgramBuildInfo(program, ezcl_get_device(context), CL_PROGRAM_BUILD_LOG, nReportSize, BuildReport, NULL);
            if (ierr != CL_SUCCESS) {
               switch (ierr){
                  case CL_INVALID_DEVICE:
                     printf("Invalid device in clProgramBuildInfo\n");
                     break;
                  case CL_INVALID_VALUE:
                     printf("Invalid value in clProgramBuildInfo\n");
                     break;
                  case CL_INVALID_PROGRAM:
                     printf("Invalid program in clProgramBuildInfo\n");
                     break;
               }
            }
              
            printf("EZCL_CREATE_KERNEL: Build Log: %s\n",BuildReport);
            free(BuildReport);
            exit(-1);
            break;
        case CL_OUT_OF_HOST_MEMORY:
          printf("Out of host memory in clBuildProgram\n");
          break;
        }
   } else {
      printf("EZCL_CREATE_PROGRAM: Build is SUCCESSFUL with no errors\n");
   }

   object_item = malloc(sizeof(struct object_entry));
   object_item->object_type = PROGRAM_OBJECT;
   object_item->program = program;
   filename_copy = (char *)malloc(strlen(filename) + 1);
   strcpy(filename_copy, filename);
   object_item->filename = filename_copy;
   if (DEBUG) printf("EZCL_CREATE_KERNEL: DEBUG -- program is %p\n",program);
   SLIST_INSERT_HEAD(&object_head, object_item, object_entries);

   kernel = clCreateKernel(program, kernel_name, &ierr);
   if (ierr != CL_SUCCESS){
      printf("EZCL_CREATE_PROGRAM: clCreateKernel returned an error %d at line %d in file %s\n",ierr,__LINE__,__FILE__);
        switch (ierr){
        case CL_INVALID_PROGRAM:
          printf("Invalid program in clCreateKernel\n");
          break;
        case CL_INVALID_PROGRAM_EXECUTABLE:
          printf("Invalid program executable in clCreateKernel\n");
          break;
        case CL_INVALID_KERNEL_NAME:
          printf("Invalid kernel name in clCreateKernel\n");
          break;
        case CL_INVALID_KERNEL_DEFINITION:
          printf("Invalid kernel definition in clCreateKernel\n");
          break;
        case CL_INVALID_VALUE:
          printf("Invalid value in clCreateKernel\n");
          break;
        case CL_OUT_OF_HOST_MEMORY:
          printf("Out of host memory in clCreateKernel\n");
          break;
        }
      exit(-1);
   }

   object_item = malloc(sizeof(struct object_entry));
   object_item->object_type = KERNEL_OBJECT;
   object_item->kernel = kernel;
   if (DEBUG) printf("EZCL_CREATE_KERNEL: DEBUG -- kernel is %p\n",kernel);
   SLIST_INSERT_HEAD(&object_head, object_item, object_entries);

   return(kernel);
}

void ezcl_enqueue_write_buffer(cl_command_queue command_queue, cl_mem mem_buffer, cl_bool blocking_flag, 
                            size_t offset, size_t size, const void *ptr, cl_event *event)
{
   int ierr;
   ierr=clEnqueueWriteBuffer(command_queue, mem_buffer, blocking_flag, offset, size, ptr, 0, NULL, event);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_COMMAND_QUEUE:
      *  CL_INVALID_CONTEXT:
      *  CL_INVALID_MEM_OBJECT:
      *  CL_INVALID_VALUE:
      *  CL_INVALID_EVENT_WAIT_LIST:
      *  CL_MEM_OBJECT_ALLOCATION_FAILURE:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_ENQUEUE_WRITE_BUFFER", "clEnqueueWriteBuffer", __LINE__, __FILE__);
   }
}

void ezcl_enqueue_read_buffer(cl_command_queue command_queue, cl_mem mem_buffer, cl_bool blocking_flag, 
                               size_t offset, size_t size, void *ptr, cl_event *event)
{
   int ierr;
   ierr=clEnqueueReadBuffer(command_queue, mem_buffer, blocking_flag, offset, size, ptr, 0, NULL, event);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_COMMAND_QUEUE:
      *  CL_INVALID_CONTEXT:
      *  CL_INVALID_MEM_OBJECT:
      *  CL_INVALID_VALUE:
      *  CL_INVALID_EVENT_WAIT_LIST:
      *  CL_MEM_OBJECT_ALLOCATION_FAILURE:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_ENQUEUE_READ_BUFFER", "clEnqueueReadBuffer", __LINE__, __FILE__);
   }
}

void ezcl_enqueue_ndrange_kernel(cl_command_queue command_queue, cl_kernel kernel, cl_uint work_dim, 
                                 const size_t *global_work_offset, const size_t *global_work_size, const size_t *local_work_size, cl_event *event)
{
   int ierr; 
   ierr=clEnqueueNDRangeKernel(command_queue, kernel, work_dim, global_work_offset, global_work_size, local_work_size, 0, NULL, event);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_PROGRAM_EXECUTABLE:
      *  CL_INVALID_COMMAND_QUEUE:
      *  CL_INVALID_KERNEL:
      *  CL_INVALID_CONTEXT:
      *  CL_INVALID_KERNEL_ARGS:
      *  CL_INVALID_WORK_DIMENSION:
      *  CL_INVALID_GLOBAL_WORK_SIZE:
      *  CL_INVALID_WORK_GROUP_SIZE:
      *  CL_INVALID_WORK_ITEM_SIZE:
      *  CL_INVALID_GLOBAL_OFFSET:
      *  CL_OUT_OF_RESOURCES:
      *  CL_MEM_OBJECT_ALLOCATION_FAILURE:
      *  CL_INVALID_EVENT_WAIT_LIST:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_ENQUEUE_NDRANGE_KERNEL", "clEnqueueNDRangeKernel", __LINE__, __FILE__);
   }
}

void ezcl_set_kernel_arg(cl_kernel kernel, cl_int arg_index, size_t arg_size, const void *arg_value)
{
   int ierr;
   ierr=clSetKernelArg(kernel, arg_index, arg_size, arg_value);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_KERNEL:
      *  CL_INVALID_ARG_INDEX:
      *  CL_INVALID_ARG_VALUE:
      *  CL_INVALID_MEM_OBJECT:
      *  CL_INVALID_SAMPLER:
      *  CL_INVALID_ARG_SIZE:
      */
     ezcl_print_error(ierr, "EZCL_SET_KERNEL_ARG", "clSetKernelArg", __LINE__, __FILE__);
   }
}

void ezcl_flush(cl_command_queue command_queue)
{
   int ierr;
   ierr=clFlush(command_queue);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_INVALID_COMMAND_QUEUE:
      *  CL_OUT_OF_HOST_MEMORY:
      */
     ezcl_print_error(ierr, "EZCL_FLUSH", "clFlush", __LINE__, __FILE__);
   }
}

void ezcl_get_event_profiling_info(cl_event event, cl_profiling_info param_name, size_t param_value_size, void *param_value)
{
   int ierr;
   ierr=clGetEventProfilingInfo(event, param_name, param_value_size, param_value, NULL);
   if (ierr != CL_SUCCESS) {
     /* Possible Errors
      *  CL_PROFILING_INFO_NOT_AVAILABLE:
      *  CL_INVALID_VALUE:
      *  CL_INVALID_EVENT:
      */
     ezcl_print_error(ierr, "EZCL_GET_EVENT_PROFILING_INFO", "clGetEventProfilingInfo", __LINE__, __FILE__);
   }
}

void ezcl_set_timing_flag(){
   ezcl_flags.timing = 1;
}

cl_int ezcl_finish(void){

   ezcl_mem_free_all();
   
   return(0);
}

void ezcl_print_error(const int ierr, const char *routine, const char *cl_routine, const int line, const char *file)
{
   switch (ierr){
      case CL_DEVICE_NOT_FOUND:                //#define CL_DEVICE_NOT_FOUND                 -1
         printf("ERROR: %s -- Device not found in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_DEVICE_NOT_AVAILABLE:            //#define CL_DEVICE_NOT_AVAILABLE             -2
         printf("ERROR: %s -- Device not available in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_COMPILER_NOT_AVAILABLE:          //#define CL_COMPILER_NOT_AVAILABLE           -3
         printf("ERROR: %s -- CL compiler not available failure in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_MEM_OBJECT_ALLOCATION_FAILURE:   //#define CL_MEM_OBJECT_ALLOCATION_FAILURE    -4
         printf("ERROR: %s -- Mem object allocation failure in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_OUT_OF_RESOURCES:                //#define CL_OUT_OF_RESOURCES                 -5
         printf("ERROR: %s -- Out of resources in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_OUT_OF_HOST_MEMORY:              //#define CL_OUT_OF_HOST_MEMORY               -6
         printf("ERROR: %s -- Out of host memory in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_PROFILING_INFO_NOT_AVAILABLE:    //#define CL_PROFILING_INFO_NOT_AVAILABLE     -7
         printf("ERROR: %s -- Profiling info not available in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_MEM_COPY_OVERLAP:                //#define CL_MEM_COPY_OVERLAP                 -8
         printf("ERROR: %s -- Mem copy overlap in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_IMAGE_FORMAT_MISMATCH:           //#define CL_IMAGE_FORMAT_MISMATCH            -9
         printf("ERROR: %s -- Image format mismatch in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_IMAGE_FORMAT_NOT_SUPPORTED:      //#define CL_IMAGE_FORMAT_NOT_SUPPORTED      -10
         printf("ERROR: %s -- Image format not supported in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_BUILD_PROGRAM_FAILURE:           //#define CL_BUILD_PROGRAM_FAILURE           -11
         printf("ERROR: %s -- Build program failure in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_MAP_FAILURE:                     //#define CL_MAP_FAILURE                     -12
         printf("ERROR: %s -- Map failure in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;

      case CL_INVALID_VALUE:                   //#define CL_INVALID_VALUE                   -30
         printf("ERROR: %s -- Invalid value in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_DEVICE_TYPE:             //#define CL_INVALID_DEVICE_TYPE             -31
         printf("ERROR: %s -- Invalid device type in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_PLATFORM:                //#define CL_INVALID_PLATFORM                -32
         printf("ERROR: %s -- Invalid platform in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_DEVICE:                  //#define CL_INVALID_DEVICE                  -33
         printf("ERROR: %s -- Invalid device in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_CONTEXT:                 //#define CL_INVALID_CONTEXT                 -34
         printf("ERROR: %s -- Invalid context in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_QUEUE_PROPERTIES:        //#define CL_INVALID_QUEUE_PROPERTIES        -35
         printf("ERROR: %s -- Invalid queue properties in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_COMMAND_QUEUE:           //#define CL_INVALID_COMMAND_QUEUE           -36
         printf("ERROR: %s -- Invalid command queue in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_HOST_PTR:                //#define CL_INVALID_HOST_PTR                -37
         printf("ERROR: %s -- Invalid host pointer in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_MEM_OBJECT:              //#define CL_INVALID_MEM_OBJECT              -38
         printf("ERROR: %s -- Invalid memory object in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR: //#define CL_INVALID_IMAGE_FORMAT_DESCRIPTOR -39
         printf("ERROR: %s -- Invalid image format descriptor in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_IMAGE_SIZE:              //#define CL_INVALID_IMAGE_SIZE              -40
         printf("ERROR: %s -- Invalid image size in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_SAMPLER:                 //#define CL_INVALID_SAMPLER                 -41
         printf("ERROR: %s -- Invalid sampler in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_BINARY:                  //#define CL_INVALID_BINARY                  -42
         printf("ERROR: %s -- Invalid binary in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_BUILD_OPTIONS:           //#define CL_INVALID_BUILD_OPTIONS           -43
         printf("ERROR: %s -- Invalid build options in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_PROGRAM:                 //#define CL_INVALID_PROGRAM                 -44
         printf("ERROR: %s -- Invalid program in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_PROGRAM_EXECUTABLE:      //#define CL_INVALID_PROGRAM_EXECUTABLE      -45
         printf("ERROR: %s -- Invalid program executable in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_KERNEL_NAME:             //#define CL_INVALID_KERNEL_NAME             -46
         printf("ERROR: %s -- Invalid kernel name in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_KERNEL_DEFINITION:       //#define CL_INVALID_KERNEL_DEFINITION       -47
         printf("ERROR: %s -- Invalid kernel definition in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_KERNEL:                  //#define CL_INVALID_KERNEL                  -48
         printf("ERROR: %s -- Invalid kernel in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_ARG_INDEX:               //#define CL_INVALID_ARG_INDEX               -49
         printf("ERROR: %s -- Invalid arg index in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_ARG_VALUE:               //#define CL_INVALID_ARG_VALUE               -50
         printf("ERROR: %s -- Invalid arg value in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_ARG_SIZE:                //#define CL_INVALID_ARG_SIZE                -51
         printf("ERROR: %s -- Invalid arg size in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_KERNEL_ARGS:             //#define CL_INVALID_KERNEL_ARGS             -52
         printf("ERROR: %s -- Invalid kernel args in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_WORK_DIMENSION:          //#define CL_INVALID_WORK_DIMENSION          -53
         printf("ERROR: %s -- Invalid work dimension in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_WORK_GROUP_SIZE:         //#define CL_INVALID_WORK_GROUP_SIZE         -54
         printf("ERROR: %s -- Invalid work group size in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_WORK_ITEM_SIZE:          //#define CL_INVALID_WORK_ITEM_SIZE          -55
         printf("ERROR: %s -- Invalid work item size in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_GLOBAL_OFFSET:           //#define CL_INVALID_GLOBAL_OFFSET           -56
         printf("ERROR: %s -- Invalid global offset in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_EVENT_WAIT_LIST:         //#define CL_INVALID_EVENT_WAIT_LIST         -57
         printf("ERROR: %s -- Invalid event wait list in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_EVENT:                   //#define CL_INVALID_EVENT                   -58
         printf("ERROR: %s -- Invalid event in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_OPERATION:               //#define CL_INVALID_OPERATION               -59
         printf("ERROR: %s -- Invalid operation in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_GL_OBJECT:               //#define CL_INVALID_GL_OBJECT               -60
         printf("ERROR: %s -- Invalid GL object in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_BUFFER_SIZE:             //#define CL_INVALID_BUFFER_SIZE             -61
         printf("ERROR: %s -- Invalid buffer size in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_MIP_LEVEL:               //#define CL_INVALID_MIP_LEVEL               -62
         printf("ERROR: %s -- Invalid mip level in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;
      case CL_INVALID_GLOBAL_WORK_SIZE:        //#define CL_INVALID_GLOBAL_WORK_SIZE        -63
         printf("ERROR: %s -- Invalid global work size in %s at line %d in file %s\n", routine, cl_routine, line, file);
         break;

      default:
        printf("ERROR: %s -- %d in %s at line %d in file %s\n", routine, ierr, cl_routine, line, file);
   }
   exit(-1);
}

void *genvector(int inum, size_t elsize)
{
   void *out;
   int mem_size;

   mem_size = inum*elsize;
   out      = (void *)calloc(inum, elsize);
   ezcl_malloc_memory_add(out, mem_size);

   return (out);
}

void genvectorfree(void *var)
{
   ezcl_malloc_memory_remove(var);
}

void **genmatrix(int jnum, int inum, size_t elsize)
{
   void **out;
   int mem_size;
  
   mem_size = jnum*sizeof(void *);
   out      = (void **)malloc(mem_size);
   ezcl_malloc_memory_add(out, mem_size);
  
   mem_size = jnum*inum*elsize;
   out[0]    = (void *)calloc(jnum*inum, elsize);
   ezcl_malloc_memory_add(out[0], mem_size);
  
   for (int i = 1; i < jnum; i++) {
      out[i] = out[i-1] + inum*elsize;
   }
  
   return (out);
}

void genmatrixfree(void **var)
{
   ezcl_malloc_memory_remove(var[0]);
   ezcl_malloc_memory_remove(var);
}

void ***gentrimatrix(int knum, int jnum, int inum, size_t elsize)
{
   void ***out;
   int mem_size;

   mem_size  = knum*sizeof(void **);
   out       = (void ***)malloc(mem_size);
   ezcl_malloc_memory_add(out, mem_size);

   mem_size  = knum*jnum*sizeof(void *);
   out[0]    = (void **) malloc(mem_size);
   ezcl_malloc_memory_add(out[0], mem_size);

   mem_size  = knum*jnum*inum*elsize;
   out[0][0] = (void *)calloc(knum*jnum*inum, elsize);
   ezcl_malloc_memory_add(out[0][0], mem_size);

   for (int k = 0; k < knum; k++)
   {
      if (k > 0)
      {
         out[k] = out[k-1] + jnum*sizeof(void *);
         out[k][0] = out[k-1][0] + (jnum*inum);
      }

      for (int j = 1; j < jnum; j++)
      {
         out[k][j] = out[k][j-1] + inum*elsize;
      }
   }
  
   return (out);
}

void gentrimatrixfree(void ***var)
{
   ezcl_malloc_memory_remove(var[0][0]);
   ezcl_malloc_memory_remove(var[0]);
   ezcl_malloc_memory_remove(var);
}


