/*
 *  ezcl.h
 *
 *  Copyright 2010 Bob Robey, brobey@lanl.gov . All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 */

#ifndef EZCL_H_
#define EZCL_H_

#define HAVE_OPENCL 1

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_OPENCL

#ifdef __APPLE_CC__
#include <OpenCL/OpenCL.h>
#else
#include <CL/cl.h>
#endif

#ifdef HAVE_CL_DOUBLE
typedef double real;
typedef cl_double cl_real;
typedef cl_double4 cl_real4;
#else
typedef float real;
typedef cl_float cl_real;
typedef cl_float4 cl_real4;
#endif

#define PROGRAM_OBJECT       1
#define KERNEL_OBJECT        2
#define COMMAND_QUEUE_OBJECT 3
#define CONTEXT_OBJECT       4

#define EZCL_SUCCESS         0
#define EZCL_NODEVICE        1

#define EZCL_PINNED_MEMORY   1

/* init and finish routines -- use either ezcl_init or ezcl_devtype_init */
/*   the devtype_init is the simpler and recommended path */
cl_int ezcl_init(cl_context *ezcl_gpu_context, cl_context *ezcl_cpu_context, cl_context *ezcl_accelerator_context);
cl_int ezcl_devtype_init(cl_device_type device_type, cl_context *return_context, cl_command_queue *return_command_queue);
cl_int ezcl_finish(void);

/* Error reporting */
void ezcl_print_error(const int ierr, const char *routine, const char *cl_routine, const int line, const char *file);

/* device based routines */
cl_device_id ezcl_get_device(cl_context context);
void ezcl_device_info(cl_device_id device);

/* context based routines*/
cl_command_queue ezcl_create_command_queue(cl_context context);
cl_command_queue ezcl_get_command_queue(void);
cl_context ezcl_get_context(void);

/* memory routines */
cl_mem ezcl_malloc(void *host_mem_ptr, size_t dims[], size_t elsize, int flags, int ezcl_flags);
void *ezcl_malloc_memory_add(void *malloc_mem_ptr, size_t size);
void ezcl_device_memory_add(cl_mem dev_mem_ptr, size_t size);
void ezcl_mapped_memory_add(cl_mem map_mem_ptr, size_t size);
void ezcl_mem_free_all(void);
void *genvector(int inum, size_t elsize);
void genvectorfree(void *var);
void **genmatrix(int jnum, int inum, size_t elsize);
void ***gentrimatrix(int knum, int jnum, int inum, size_t elsize);
void genmatrixfree(void **var);
void gentrimatrixfree(void ***var);

/* kernel and program routines */
cl_kernel ezcl_create_kernel(cl_context context, const char *filename, const char *kernel_name, int flags);
//cl_command_queue ezcl_create_command_queue(void);

void ezcl_program_release(cl_program program);
void ezcl_kernel_release(cl_kernel kernel);
void ezcl_command_queue_release(cl_command_queue command_queue);
void ezcl_context_release(cl_context context);

void ezcl_enqueue_write_buffer(cl_command_queue command_queue, cl_mem mem_buffer, cl_bool blocking_flag,
                               size_t offset, size_t size, const void *ptr, cl_event *event);
void ezcl_enqueue_read_buffer(cl_command_queue command_queue, cl_mem mem_buffer, cl_bool blocking_flag,
                              size_t offset, size_t size, void *ptr, cl_event *event);
void ezcl_enqueue_ndrange_kernel(cl_command_queue command_queue, cl_kernel kernel, cl_uint work_dim,
                                 const size_t *global_work_offset, const size_t *global_work_size, const size_t *local_work_size, cl_event *event);
void ezcl_set_kernel_arg(cl_kernel kernel, cl_int arg_index, size_t arg_size, const void *arg_value);
void ezcl_flush(cl_command_queue command_queue);
void ezcl_get_event_profiling_info(cl_event event, cl_profiling_info param_name, size_t param_value_size, void *param_value);

void ezcl_set_timing_flag(void);

#endif
#endif
