/*
 *  memory.c
 *
 *  Copyright 2007-2010 Bob Robey, brobey@lanl.gov . All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 */

#ifndef DEBUG
#define DEBUG 0
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/queue.h>

#include "memory.h"

static SLIST_HEAD(slist_memory_head, memory_entry) memory_head = SLIST_HEAD_INITIALIZER(memory_head);
//static struct slist_memory_head *memory_headp;
static struct memory_entry {
   void *mem_ptr;
   size_t mem_size;
   SLIST_ENTRY(memory_entry) memory_entries;
} *memory_item;

double *dvector(int isize)
{
   double *out;
   int mem_size;
   
   mem_size = isize*sizeof(double);
   out      = (double *)calloc(isize, sizeof(double));
   memory_add(out, mem_size);
      
   return (out);
}

void dvectorfree(double *var)
{
   memory_free(var);
}

double **dmatrix(int jsize, int isize)
{
   double **out;
   int mem_size;
   
   mem_size = jsize*sizeof(double *);
   out      = (double **)malloc(mem_size);
   memory_add(out, mem_size);
   
   mem_size = jsize*isize*sizeof(double);
   out[0]    = (double *)calloc(jsize*isize, sizeof(double));
   memory_add(out[0], mem_size);
   
   for (int i = 1; i < jsize; i++) {
      out[i] = out[i-1] + isize;
   }
   
   return (out);
}

void dmatrixfree(double **var)
{
   memory_free(var[0]);
   memory_free(var);
}

double ***dtrimatrix(int ksize, int jsize, int isize)
{
   double ***out;
   int mem_size;
   
   mem_size  = ksize*sizeof(double **);
   out       = (double ***)malloc(mem_size);
   memory_add(out, mem_size);
   
   mem_size  = ksize*jsize*sizeof(double *);
   out[0]    = (double **) malloc(mem_size);
   memory_add(out[0], mem_size);
   
   mem_size  = ksize*jsize*isize*sizeof(double);
   out[0][0] = (double *)calloc(ksize*jsize*isize, sizeof(double));
   memory_add(out[0][0], mem_size);
   
   for (int k = 0; k < ksize; k++)
   {
      if (k > 0)
      {
         out[k] = out[k-1] + jsize;
         out[k][0] = out[k-1][0] + (jsize*isize);
      }
      
      for (int j = 1; j < jsize; j++)
      {
         out[k][j] = out[k][j-1] + isize;
      }
    }
   
   return (out);
}

void dtrimatrixfree(double ***var)
{
   memory_free(var[0][0]);   
   memory_free(var[0]);
   memory_free(var);
}

float *fvector(int isize)
{
   float *out;
   int mem_size;
   
   mem_size = isize*sizeof(float);
   out      = (float *)calloc(isize, sizeof(float));
   memory_add(out, mem_size);
   
   return (out);
}

void fvectorfree(float *var)
{
   memory_free(var);
}

float **fmatrix(int jsize, int isize)
{
   float **out;
   int mem_size;
   
   mem_size = jsize*sizeof(float *);
   out      = (float **)malloc(mem_size);
   memory_add(out, mem_size);
   
   mem_size = jsize*isize*sizeof(float);
   out[0]    = (float *)calloc(jsize*isize, sizeof(float));
   memory_add(out[0], mem_size);
   
   for (int i = 1; i < jsize; i++) {
      out[i] = out[i-1] + isize;
   }
   
   return (out);
}

void fmatrixfree(float **var)
{
   memory_free(var[0]);
   memory_free(var);
}

float ***ftrimatrix(int ksize, int jsize, int isize)
{
   float ***out;
   int mem_size;
   
   mem_size  = ksize*sizeof(float **);
   out       = (float ***)malloc(mem_size);
   memory_add(out, mem_size);
   
   mem_size  = ksize*jsize*sizeof(float *);
   out[0]    = (float **) malloc(mem_size);
   memory_add(out[0], mem_size);
   
   mem_size  = ksize*jsize*isize*sizeof(float);
   out[0][0] = (float *)calloc(ksize*jsize*isize, sizeof(float));
   memory_add(out[0][0], mem_size);
   
   for (int k = 0; k < ksize; k++)
   {
      if (k > 0)
      {
         out[k] = out[k-1] + jsize;
         out[k][0] = out[k-1][0] + (jsize*isize);
      }
      
      for (int j = 1; j < jsize; j++)
      {
         out[k][j] = out[k][j-1] + isize;
      }
   }
   
   return (out);
}

void ftrimatrixfree(float ***var)
{
   memory_free(var[0][0]);   
   memory_free(var[0]);
   memory_free(var);
}

int *ivector(int isize)
{
   int *out;
   int mem_size;
   
   mem_size = isize*sizeof(int);
   out      = (int *)calloc(isize, sizeof(int));
   memory_add(out, mem_size);
   
   return (out);
}

void ivectorfree(int *var)
{
   memory_free(var);
}

int **imatrix(int jsize, int isize)
{
   int **out;
   int mem_size;
   
   mem_size = jsize*sizeof(int *);
   out      = (int **)malloc(mem_size);
   memory_add(out, mem_size);
   
   mem_size = jsize*isize*sizeof(int);
   out[0]    = (int *)calloc(jsize*isize, sizeof(int));
   memory_add(out[0], mem_size);
   
   for (int i = 1; i < jsize; i++) {
      out[i] = out[i-1] + isize;
   }
   
   return (out);
}

void imatrixfree(int **var)
{
   memory_free(var[0]);
   memory_free(var);
}

int ***itrimatrix(int ksize, int jsize, int isize)
{
   int ***out;
   int mem_size;
   
   mem_size  = ksize*sizeof(int **);
   out       = (int ***)malloc(mem_size);
   memory_add(out, mem_size);
   
   mem_size  = ksize*jsize*sizeof(int *);
   out[0]    = (int **)malloc(mem_size);
   memory_add(out[0], mem_size);
   
   mem_size  = ksize*jsize*isize*sizeof(int);
   out[0][0] = (int *)calloc(ksize*jsize*isize, sizeof(int));
   memory_add(out[0][0], mem_size);
   
   for (int k = 0; k < ksize; k++)
   {
      if (k > 0)
      {
         out[k] = out[k-1] + jsize;
         out[k][0] = out[k-1][0] + (jsize*isize);
      }
      
      for (int j = 1; j < jsize; j++)
      {
         out[k][j] = out[k][j-1] + isize;
      }
   }
   
   return (out);
}

void itrimatrixfree(int ***var)
{
   memory_free(var[0][0]);   
   memory_free(var[0]);
   memory_free(var);
}

void memory_add(void *mem_ptr, int isize){
   if (SLIST_EMPTY(&memory_head)) SLIST_INIT(&memory_head);
      
   memory_item = malloc(sizeof(struct memory_entry));
   memory_item->mem_ptr  = mem_ptr;
   memory_item->mem_size = isize;
   if (DEBUG) printf("MEMORY: memory add -- adding memory pointer %p\n",mem_ptr);
   SLIST_INSERT_HEAD(&memory_head, memory_item, memory_entries);
}

void memory_free(void *mem_ptr){
   SLIST_FOREACH(memory_item, &memory_head, memory_entries){
      if (memory_item->mem_ptr == mem_ptr) {
         if (DEBUG) printf("MEMORY: memory free -- freeing memory pointer %p\n",mem_ptr);
         free(mem_ptr);
         SLIST_REMOVE(&memory_head, memory_item, memory_entry, memory_entries);
         free(memory_item);
            break;         
      }
   }
}

void memory_free_all(void){
   while (!SLIST_EMPTY(&memory_head)) {
      memory_item = SLIST_FIRST(&memory_head);
      if (DEBUG) printf("MEMORY: memory free all -- freeing memory %p\n",memory_item->mem_ptr);
      free(memory_item->mem_ptr);
      SLIST_REMOVE_HEAD(&memory_head, memory_entries);
      free(memory_item);
   }
}

size_t memory_report(void){
   size_t sum=0;
   SLIST_FOREACH(memory_item, &memory_head, memory_entries){
      if (DEBUG) printf("MEMORY: memory report -- summing up memory size %ld for memory pointer %p\n",memory_item->mem_size, memory_item->mem_ptr);
      sum += memory_item->mem_size + sizeof(memory_item);
   }
   return(sum);
}
