/*
 *  mesh.c
 *
 *  Copyright 2007-2010 Jon Robey and Bob Robey, brobey@lanl.gov 
 *  All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 */

#include <stdlib.h>
#include <stdio.h>
#ifdef __APPLE_CC__
#include <OpenCL/OpenCL.h>
#else
#include <CL/cl.h>
#endif
#include "mesh.h"
#include "memory.h"

#ifdef __APPLE_CC__
#define Hin(j,i)     state[j][i][0]
#define Uvin(j,i)     state[j][i][1]
#define Vin(j,i)     state[j][i][2]
#define Win(j,i)     state[j][i][3]
#define Speedin(j,i) state[j][i][3]
#else
#define Hin(j,i)     state[j][i].s0
#define Uvin(j,i)     state[j][i].s1
#define Vin(j,i)     state[j][i].s2
#define Win(j,i)     state[j][i].s3
#define Speedin(j,i) state[j][i].s3
#endif

static unsigned int isize = -1;
static unsigned int jsize = -1;
static unsigned int myoffset = 0;
static unsigned int isize_padded = -1;
static double deltaX = 0.0;
static double deltaY = 0.0;
static double xmin = 0.0;
static double ymin = 0.0;
static double xmax = 100.0;
static double ymax = 100.0;

void mesh_init(unsigned int problem, unsigned int imax, unsigned int jmax,
            unsigned int imax_padded, cl_real4** state)
{
   unsigned int i, j;
   isize = imax;
   jsize = jmax;
   isize_padded = imax_padded;
   
   // Set initial conditions of a wave starting at height 10 and sloping to 2
   // midway through the x dimension
   if (problem == SLOPED_WAVE){
      for(j=0;j<jmax;j++){
         for(i=0;i<imax;i++){
            Hin(j,i)= 2.0f;
            Uvin(j,i)= 0.0f;
            Vin(j,i)= 0.0f;
            Win(j,i)= i;
         }
      }
      
      for(j=0;j<jmax;j++){
         for(i=0;i<imax/2;i++){
            Hin(j,i)=(real)(10.0 - ((10.0 - 2.0)/ (double)(imax/2))*(double)(i));
         }
      }
      deltaX = 1.0;
      deltaY = 1.0;
   }
   if (problem == WAVE_BOX){
      for(j=0;j<jmax;j++){
         for(i=0;i<imax;i++){
            Hin(j,i)= 1.0f;
            Uvin(j,i)= 0.0f;
            Vin(j,i)= 0.0f;
            if (j+myoffset >= jsize/2 - jsize/75 -1 && j+myoffset <= jsize/2 + jsize/75 -1 &&
               i >= isize/2 - isize/75 -1 && i <= isize/2 + isize/75 -1 ){
               Hin(j,i)=10.0f;
            }
         }
      }
      deltaX = 1.0;
      deltaY = 1.0;
   }
   xmax = xmin + ((double)isize+1)*deltaX;
   ymax = ymin + ((double)jsize+1)*deltaY;
}

double mesh_get_deltaX(void) { return(deltaX); }
double mesh_get_deltaY(void) { return(deltaY); }
double mesh_get_xmin(void) { return(xmin); }
double mesh_get_ymin(void) { return(ymin); }
double mesh_get_xmax(void) { return(xmax); }
double mesh_get_ymax(void) { return(ymax); }

double *mesh_get_x(void){
   double *x=dvector(isize);
   for (int i=0; i<isize;i++){
      x[i] = xmin + ((double)i+0.5)*deltaX;
   }
   return(x);
}

double *mesh_get_dx(void){
   double *dx=dvector(isize);
   for (int i=0; i<isize;i++){
      dx[i] = deltaX;
   }
   return(dx);
}

double *mesh_get_y(void){
   double *y=dvector(jsize);
   for (int j=0; j<jsize;j++){
      y[j] = ymin + ((double)j+0.5)*deltaY;
   }
   return(y);
}

double *mesh_get_dy(void){
   double *dy=dvector(jsize);
   for (int j=0; j<jsize;j++){
      dy[j] = deltaY;
   }
   return(dy);
}
      
void mesh_mem_zero_init(int jsize, int isize, cl_real4** state)
{
   for(unsigned int j=0; j<jsize; j++){
      for(unsigned int i=0; i<isize; i++){
         Hin(j,i) = 0.0f; // H
         Uvin(j,i) = 0.0f; // U
         Vin(j,i) = 0.0f; // V
         Win(j,i) = 0.0f; // -- not used
      }
   }
}


double mesh_sum_var(cl_real4** state)
{
   double sum, correction, corrected_next_term, new_sum;
   
   // Kahan summation for more precision -- see Wikipedia or Dr. Dobbs article
   
   sum=0.0;
   correction=0.0;
   for(unsigned int j=0; j<jsize; j++){
      for(unsigned int i=0; i<isize; i++){
         corrected_next_term = Hin(j,i)*deltaX*deltaY - correction;
         new_sum = sum + corrected_next_term;
         correction = (new_sum - sum) - corrected_next_term;
         sum = new_sum;
      }
   }
   
   return(sum);
}

void mesh_set_plot_var(cl_real4** state, double **plot_var)
{
   for(unsigned int j=0; j<jsize; j++){
      for(unsigned int i=0; i<isize; i++){
         plot_var[j][i]=Hin(j,i);
      }
   }
}

double mesh_set_timestep(double sigma, cl_real4** state)
{
   double curr, deltaT;
   
   double localmax=0.0;
   
   for(unsigned int j=0; j<jsize; j++){
      for(unsigned int i=0; i<isize; i++){
         
         curr=Speedin(j,i);
         if(curr>localmax) localmax=curr;
      }
   }
   
   deltaT=sigma/localmax;
   
   return(deltaT);
}


void mesh_output(char *title, cl_real4** state)
{
    printf("\n   %s\n\n",title);
    if (isize > 16){
      for (unsigned int j=0;j<jsize;j++){
         for (unsigned int i=0;i<isize;i++){
            printf("%s -- j %d i %d H %f U %f V %f\n",title,j,i,
                  Hin(j,i),Uvin(j,i),Vin(j,i));
         }
      }
    }
    else {
      printf("%s -- col     ",title);
      for (unsigned int i=0;i<isize;i++){
         printf(" %6d ",i);
      }
      printf("\n");
      
      //for (unsigned int j=0;j<jsize;j++){
      for (unsigned int j=0;j<1;j++){
         
         printf("%s -- row %3d H ",title,j);
         for (unsigned int i=0;i<isize;i++){
            printf(" %6.3f ",Hin(j,i));
         }
         printf("\n");
         
         printf("%s --         U ",title,j);
         for (unsigned int i=0;i<isize;i++){
            printf(" %6.3f ",Uvin(j,i));
         }
         printf("\n");
         
         printf("%s --         V ",title,j);
         for (unsigned int i=0;i<isize;i++){
            printf(" %6.3f ",Vin(j,i));
         }
         printf("\n");
      }
    }
   
}
