/*********************************************************************************************
 * Modified for the GPU and adding TVD based on Sapient
 *  Copyright 2010 Bob Robey, brobey@lanl.gov . All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 *
 * WAVE -- 2D Shallow Water Equation Model
 *                            Bob Robey, Los Alamos National Laboratory
 *                      Written for Supercomputing Challenge Kickoff 2007
 *                                   Copyright, 2007
 *                      Permission Granted to Supercomputing Challenge
 *                           Participants to use in their projects
 *
 * This work is based on the following 2006-2007 Supercomputing
 * Challenge Projects. Their work is used by permission.
 *
 * Numerical method based on the Shallow Water Model by
 *
 *             Joseph Koby, Sarah Armstrong, Juan-Antonio Vigil, Vanessa Trujillo,
 *                                  McCurdy High School
 *                                    Copyright 2007
 *                        Written for Supercomputing Challenge 2006-2007
 *                                ALL RIGHTS RESERVED
 *
 * C/MPI coding based on the SAPIENT Compressible Fluid Model
 *
 *                                      SAPIENT
 *                  Copyright 2007 Jonathan Robey, Dov Shlachter
 *                               Los Alamos High School
 *                     Written for Supercomputing Challenge 2006-2007
 *                                  ALL RIGHTS RESERVED
 *
 * ******************************************************************************************/

#include <stdio.h>

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include "ezcl.h"
#include "wave.h"
#include "mesh.h"
#include "display.h"
#include "memory.h"

#define SWAP(xnew,xold,xtmp) (xtmp=xnew, xnew=xold, xold=xtmp)
#define MIN(a,b) ((a)>(b)?(b):(a))

int main (int argc, char *argv[]) {
   // Control flags
   struct {
      unsigned int runflag    : 1;
      unsigned int timing     : 1;
   } flags={1,1};

   char string[80];
   cl_mem dev_tmp;
   int ibutton = -1;
   int special_event;
   char key;
   int xloc, yloc;
   int icell, jcell;
   double **display_var;
   double xcor, ycor;
   int imax = 1280;
   int jmax = 1280;
   int nstep = 10000;
   int nburst = 100;     // MUST divide the value of nstep without remainder
   int ndisplay = 100;  // MUST be a multiple of nburst
    
   //int meshRow;
   double time=0.0;
   double colorScale=10.0;
   double graphScale=11.0;
   struct timeval tstart, tstop, tresult;
   double TotalMass;
   double deltaX, deltaY;
   // Make local x tile size a maximum of 16 or the x dimension if smaller
   int local_size_x=MIN(imax,16); // chosen for NVIDIA GeForce 9800 GTX/9800 GTX+ w/512 MB
   int local_size_y=MIN(jmax,8);
   // Make global size an integer multiple of the local_size
   int global_size_x=((imax+local_size_x-1)/local_size_x)*local_size_x;
   int global_size_y=((jmax+local_size_y-1)/local_size_y)*local_size_y;
   size_t memsize=global_size_x*global_size_y;
   //int num_groups=global_size_x*global_size_y/(local_size_x*local_size_y);
   size_t global_work_size[2];
   size_t local_work_size[2];
    
   size_t elsize = 0;

   global_work_size[0] = global_size_x;
   global_work_size[1] = global_size_y;
    
   local_work_size[0] = local_size_x;
   local_work_size[1] = local_size_y;
   
   size_t sum_global_work_size[1];
   size_t sum_local_work_size[1];
   
   sum_global_work_size[0] = global_size_x*global_size_y;
   sum_local_work_size[0] = local_size_x*local_size_y;
    
   double deltaT = 0.025;       //initial timestep
   real deltaTf;
   int isize, ioffset, isize_old;
    
   int ierr;
   cl_event write_event, start_compute_event, end_compute_event, reduction_compute_event, read_event;
   long time_start, time_end, time_compute=0, time_read=0, time_reduction=0, time_total;
   long time_compute_burst, time_read_burst;
   long time_reduction_burst;
   cl_context context = NULL;
   cl_kernel kernel = NULL;
   cl_kernel kernel_sum = NULL;
   cl_command_queue command_queue = NULL;
   cl_real4** state_in  = NULL;
   cl_real4** state_out = NULL;
   cl_mem dev_state_in  = NULL;
   cl_mem dev_state_out = NULL;

   ierr = ezcl_devtype_init(CL_DEVICE_TYPE_GPU, &context, &command_queue);
   if (ierr == EZCL_NODEVICE){
      ierr = ezcl_devtype_init(CL_DEVICE_TYPE_CPU, &context, &command_queue);
   }
   if (ierr != EZCL_SUCCESS) {
      printf("No opencl device available -- aborting\n");
      exit(-1);      
   }
   
   if (flags.timing) ezcl_set_timing_flag();

   kernel = ezcl_create_kernel(context, "wave_kern.cl","wave_kern",0);
   kernel_sum = ezcl_create_kernel(context, "finish_sum.cl","finish_sum",0);
   
   elsize = sizeof(cl_real4);

   state_in = (cl_real4 **)genmatrix(global_size_y, global_size_x, elsize);
   state_out = (cl_real4 **)genmatrix(global_size_y, global_size_x, elsize);

   mesh_mem_zero_init(global_size_y, global_size_x, state_in);
   mesh_mem_zero_init(global_size_y, global_size_x, state_out);
    
   mesh_init(PROBLEM, imax, jmax, global_size_x, state_in);
   deltaX = mesh_get_deltaX();
   deltaY = mesh_get_deltaY();
   double xmin = mesh_get_xmin();
   double xmax = mesh_get_xmax();
   double ymin = mesh_get_ymin();
   double ymax = mesh_get_ymax();
   double *x=mesh_get_x();
   double *y=mesh_get_y();
   double *dx=mesh_get_dx();
   double *dy=mesh_get_dy();
   
   dev_state_in  = ezcl_malloc(NULL, &memsize, elsize, CL_MEM_READ_ONLY, 0);
   dev_state_out = ezcl_malloc(NULL, &memsize, elsize, CL_MEM_WRITE_ONLY, 0);

   ezcl_enqueue_write_buffer(command_queue, dev_state_in, CL_TRUE, 0, memsize*sizeof(cl_real4), (void *)state_in[0], &write_event);
    
   //mesh_output("Init", state_in);
   if (DISPLAY_ON){
      display_init(&argc, &argv, 800, 800, 600, xmin, xmax, ymin, ymax);
      //display_init(displayname, 400, 400, 200);
      display_set_colorscale(colorScale);
      display_set_graphscale(graphScale);
      display_set_graphrow(jmax/2);
      display_var = dmatrix(jmax,imax);
        
      mesh_set_plot_var(state_in, display_var);
      display_set_label("initial values of height");
      //display_set_window(0,xmax/2.0,0,ymax/2.0);
      display_data(jmax, imax, 0, jmax, display_var, x, dx, y, dy);
      sleep(WAIT_TIME);
   }
   TotalMass=mesh_sum_var(state_in);
    
   printf("Iteration:%8d, Time:%8.2f, Timestep: %.3f TotalMass: %f\n", 0, time, deltaT, TotalMass);
    
   real deltaXf = (real) deltaX;
   real deltaYf = (real) deltaY;
   ezcl_set_kernel_arg(kernel, 1, sizeof(typeof(deltaXf)), (void *)&deltaXf);
   ezcl_set_kernel_arg(kernel, 2, sizeof(typeof(deltaYf)), (void *)&deltaYf);
   ezcl_set_kernel_arg(kernel, 3, sizeof(typeof(imax)),   (void *)&imax);
   ezcl_set_kernel_arg(kernel, 4, sizeof(typeof(jmax)),   (void *)&imax);
   size_t shared_size = (local_size_x+4) * (local_size_y+4) * sizeof(cl_real4);
   ezcl_set_kernel_arg(kernel, 7, shared_size, 0);

   isize = sum_global_work_size[0]/(local_size_x*local_size_y);
   ioffset = isize;
   isize_old = sum_global_work_size[0];
   ezcl_set_kernel_arg(kernel_sum, 0, sizeof(typeof(isize)),   (void *)&isize);
   ezcl_set_kernel_arg(kernel_sum, 1, sizeof(typeof(isize_old)),   (void *)&isize_old);
   ezcl_set_kernel_arg(kernel_sum, 2, sizeof(typeof(ioffset)),   (void *)&ioffset);
   ezcl_set_kernel_arg(kernel_sum, 4, shared_size, 0);
   
   gettimeofday(&tstart, NULL);
    
   for(int istep=0; istep<nstep&&flags.runflag; ) {
      deltaTf = (real) deltaT;
      ezcl_set_kernel_arg(kernel, 0, sizeof(typeof(deltaTf)), (void *)&deltaTf);

      for(int iburst=0; iburst<nburst; iburst++) {
         ezcl_set_kernel_arg(kernel, 5, sizeof(cl_mem), (void *)&dev_state_in);
         ezcl_set_kernel_arg(kernel, 6, sizeof(cl_mem), (void *)&dev_state_out);
         if (iburst == 0){
            ezcl_enqueue_ndrange_kernel(command_queue, kernel, 2, NULL, global_work_size, local_work_size, &start_compute_event);
         } else if (iburst + 1 == nburst) {
            ezcl_enqueue_ndrange_kernel(command_queue, kernel, 2, NULL, global_work_size, local_work_size, &end_compute_event);
         } else {
            ezcl_enqueue_ndrange_kernel(command_queue, kernel, 2, NULL, global_work_size, local_work_size, NULL);
         }
         if (isize > 1){
            ezcl_set_kernel_arg(kernel_sum, 3, sizeof(cl_mem), (void *)&dev_state_out);
            //printf("DEBUG -- sum_local_work_size %ld sum_global_work_size %ld isize %d\n",sum_local_work_size[0],sum_global_work_size[0],isize);
            ezcl_enqueue_ndrange_kernel(command_queue, kernel_sum, 1, NULL, sum_local_work_size, sum_local_work_size, &reduction_compute_event);
         }
         ezcl_enqueue_read_buffer(command_queue, dev_state_out, CL_TRUE, 0, 2*sizeof(cl_real4), state_out[0], &read_event);
#ifdef __APPLE_CC__
         TotalMass = (double)state_out[0][0][3];
         deltaT    = 0.9/(double)state_out[0][1][3];
#else
         TotalMass = (double)state_out[0][0].s3;
         deltaT    = 0.9/(double)state_out[0][1].s3;
#endif
         //printf("Sum is %lf\n",sum);
         SWAP(dev_state_out, dev_state_in, dev_tmp);
      }
      ezcl_enqueue_read_buffer(command_queue, dev_state_in, CL_TRUE, 0, memsize*sizeof(cl_real4), state_in[0], &read_event);
      //ezcl_enqueue_read_buffer(command_queue, dev_state_out, CL_FALSE, 0, memsize*sizeof(cl_real4), state_out[0], NULL);
      ezcl_flush(command_queue);
      
      if (flags.timing){
         ezcl_get_event_profiling_info(start_compute_event, CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start);
         ezcl_get_event_profiling_info(end_compute_event,   CL_PROFILING_COMMAND_END,   sizeof(time_end),   &time_end);
         time_compute_burst = time_end - time_start;
         time_compute += time_compute_burst;
         
         ezcl_get_event_profiling_info(reduction_compute_event, CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start);
         ezcl_get_event_profiling_info(reduction_compute_event, CL_PROFILING_COMMAND_END,   sizeof(time_end),   &time_end);
         time_reduction_burst = time_end - time_start;
         time_reduction += time_reduction_burst;
         
         ezcl_get_event_profiling_info(read_event, CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start);
         ezcl_get_event_profiling_info(read_event, CL_PROFILING_COMMAND_END,   sizeof(time_end),   &time_end);
         time_read_burst = time_end - time_start;
         time_read += time_read_burst;
      }
      
      //mesh_output("cycle",state_in);
      istep += nburst;
      time += nburst*deltaT;
      printf("Iteration:%8d, Time:%8.2f, Timestep: %.3f TotalMass: %f  Time per Cycle: %.2f  msecs. Read from device: %.2f msecs.\n",
            istep, time, deltaT, TotalMass,
            (double)time_compute_burst * 1.0e-6/(double)nburst, (double)time_read_burst * 1.0e-6);
      if (DISPLAY_ON && istep%ndisplay == 0){
         mesh_set_plot_var(state_in, display_var);
         //mass or height
         sprintf(string,"%s     Iter %d      time %lf","height",istep,time);
         display_set_label(string);
         display_data_interactive(jmax, imax, 0, jmax, display_var, x, dx, y, dy, 
           &ibutton, &key, &xloc, &yloc, &xcor, &ycor, &icell, &jcell, &special_event);

         if (ibutton != -1 || key != '\0'){
            //xycoord_tocell_index(xcor, ycor, icol, jrow);
         }
        
         //if (ibutton || key !='\0' || special_event || icell >=0 ) {
         //   printf("INPUT -- button is %d key is '%c' special_event is %d icell is %d loc is %d %d coor is %lf %lf\n",
         //         ibutton,      key,        special_event,      icell,      xloc, yloc, xcor, ycor);
         //}
         switch (ibutton){
         case 1:
            break;
         }
         switch(special_event){
         case 1:
            break;
         }
         switch(key){
         case 'E':
            printf("\nKey E: End -- ending calculation\n\n");
            flags.runflag=0;
            break;
         case 'M':
            printf("        MENU\n");
            printf("\n  Control Keys\n");
            printf("Key E: End     -- ending calculation\n");
            printf("Key M: Menu    -- print out menu for interactive input\n");
            printf("Key Q: Quiet   -- turn off display\n");
            printf("Key V: Verbose -- turn back on display\n");
            printf("\n");
            break;
         case 'Q':
            printf("\nKey Q: Quiet -- turn off display\n\n");
            display_off();
            break;
         case 'V': 
            printf("\nKey V: Verbose -- turn back on display\n\n");
            display_on();
                  
            break;
         }
      }

   } // End of iteration loop
            
   gettimeofday(&tstop, NULL);
   timersub(&tstop,&tstart,&tresult);
    
   double elapsed_time = tresult.tv_sec + tresult.tv_usec/1000000.0;
   size_t memory_sum = memory_report();
    
   printf("Wave gpu C timer, Time = %.5f s, Memory used is %.6f Mbytes\n\n",elapsed_time, (double)memory_sum/1024.0/1024.0);

   if (flags.timing){
      ierr = clGetEventProfilingInfo(write_event, CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start, NULL);
      if (ierr != CL_SUCCESS) time_start = 0;
      ierr = clGetEventProfilingInfo(write_event, CL_PROFILING_COMMAND_END,   sizeof(time_end),   &time_end,   NULL);
      if (ierr != CL_SUCCESS) time_end = 0;
      time_total = (time_end - time_start) + time_compute + time_read;
      printf("Profiling: Write to device  time was %12.4f secs.\n", (double)(time_end - time_start) * 1.0e-9); /* Convert nanoseconds to msecs */
      printf("Profiling: Device compute   time was %12.4f secs.\n", (double) time_compute           * 1.0e-9);
      printf("Profiling: Reduction        time was %12.4f secs.\n", (double) time_reduction         * 1.0e-9);
      printf("Profiling: Read from device time was %12.4f secs.\n", (double) time_read              * 1.0e-9);
      printf("----------------------------------------------------\n");
      printf("Profiling: Total            time was %12.4f secs  or %8.2f mins.\n",
             (double) time_total * 1.0e-9, (double) time_total * 1.0e-9 /60.0);
   }

   //mesh_output("End",state_in);
    
   memory_free_all(); /* free memory from memory package */
   ezcl_finish();     /* free memory from ezcl package and close */

   if(DISPLAY_ON){
      usleep((10+WAIT_TIME)*100000);
      display_close();
   }

   return 0;
}
